module dual_core
  (
    // Outputs
    out_ready, out_result, out_valid, out_init_load, out_init_addr, out_init_data,
    // Inputs
    clk, reset_n, inp_request, inp_valid, inp_ready, in_init_load, in_init_addr, in_init_data,
    // Debug
    active_threads
  );


  import types_pkg::*;


  // Control
  input  logic                                reset_n;
  input  logic                                clk;

  // Request interface
  input  logic [REQUEST_BITS            -1:0] inp_request;
  input  logic                                inp_valid;
  output logic                                out_ready;

  // Result interface
  output logic [RESULT_BITS             -1:0] out_result;
  output logic                                out_valid;
  input  logic                                inp_ready;

  // Object buffer input initialisation
  input  logic                                in_init_load;
  input  logic [OBJ_BUFFER_ADDR_BITS    -1:0] in_init_addr;
  input  logic [WORD_BITS               -1:0] in_init_data;

  // Object buffer output initialisation
  output logic                                out_init_load;
  output logic [OBJ_BUFFER_ADDR_BITS    -1:0] out_init_addr;
  output logic [WORD_BITS               -1:0] out_init_data;

  // Debug interface
  output integer                              active_threads;



  // Core 0 request interface
  logic                                       c0_inp_valid;
  logic                                       c0_out_ready;
  
  // Core 0 result interface
  logic                                       c0_out_valid;
  logic                                       c0_inp_ready;
  logic [RESULT_BITS                    -1:0] c0_out_result;

  // Core 0 object buffer interface
  logic [OBJ_BUFFER_ADDR_BITS           -1:0] c0_obj_addr;
  logic                                       c0_obj_read;
  logic [WORD_BITS                      -1:0] c0_obj_data;

  // Core 0 instruction memory interface
  logic [PC_BITS                        -1:0] c0_instruction_address;
  instruction_union_t                         c0_instruction;

  // Core 0 debug interface
  integer                                     c0_active_threads;


  // Core 1 request interface
  logic                                       c1_inp_valid;
  logic                                       c1_out_ready;

  // Core 1 result interface
  logic                                       c1_out_valid;
  logic                                       c1_inp_ready;
  logic [RESULT_BITS                    -1:0] c1_out_result;
  
  // Core 1 object buffer interface
  logic [OBJ_BUFFER_ADDR_BITS           -1:0] c1_obj_addr;
  logic                                       c1_obj_read;
  logic [WORD_BITS                      -1:0] c1_obj_data;

  // Core 1 instruction memory interface
  logic [PC_BITS                        -1:0] c1_instruction_address;
  instruction_union_t                         c1_instruction;

  // Core 1 debug interface
  integer                                     c1_active_threads;

  assign out_ready = c0_out_ready | c1_out_ready;

  assign c0_inp_valid = inp_valid & c0_out_ready;
  assign c1_inp_valid = inp_valid & c1_out_ready & ~c0_out_ready;

  assign out_valid = c0_out_valid | c1_out_valid;

  assign c0_inp_ready = inp_ready & c0_out_valid;
  assign c1_inp_ready = inp_ready & c1_out_valid & ~c0_out_valid;

  assign out_result = ( c1_inp_ready ? c1_out_result : c0_out_result );


  assign active_threads = c0_active_threads + c1_active_threads;
   

  ray_core C0 (
    .reset_n(reset_n),
    .clk(clk),
    .inp_request(inp_request),
    .inp_valid(c0_inp_valid),
    .out_ready(c0_out_ready),
    .out_result(c0_out_result),
    .out_valid(c0_out_valid),
    .inp_ready(c0_inp_ready),
    .out_obj_addr(c0_obj_addr),
    .out_obj_read(c0_obj_read),
    .inp_obj_data(c0_obj_data),
    .out_instruction_address(c0_instruction_address),
    .inp_instruction(c0_instruction),
    .active_threads(c0_active_threads)
  );

  ray_core C1 (
    .reset_n(reset_n),
    .clk(clk),
    .inp_request(inp_request),
    .inp_valid(c1_inp_valid),
    .out_ready(c1_out_ready),
    .out_result(c1_out_result),
    .out_valid(c1_out_valid),
    .inp_ready(c1_inp_ready),
    .out_obj_addr(c1_obj_addr),
    .out_obj_read(c1_obj_read),
    .inp_obj_data(c1_obj_data),
    .out_instruction_address(c1_instruction_address),
    .inp_instruction(c1_instruction),
    .active_threads(c1_active_threads)
  );

  instruction_memory IMEM (
    .*
  );

  object_buffer OBUF (
    .*
  );


  // bind ray_core ray_core_property U0 (.*);



endmodule