module dual_core_array
  (
    // Outputs
    out_ready, out_result, out_valid,
    // Inputs
    clk, reset_n, inp_request, inp_valid, inp_ready, in_init_load, in_init_addr, in_init_data,
    // Debug
    active_threads
  );


  import types_pkg::*;

  parameter NUM_DCORES = 2;


  // Control
  input  logic                                reset_n;
  input  logic                                clk;

  // Request interface
  input  logic [REQUEST_BITS            -1:0] inp_request;
  input  logic                                inp_valid;
  output logic                                out_ready;

  // Result interface
  output logic [RESULT_BITS             -1:0] out_result;
  output logic                                out_valid;
  input  logic                                inp_ready;

  // Object buffer input initialisation
  input  logic                                in_init_load;
  input  logic [OBJ_BUFFER_ADDR_BITS    -1:0] in_init_addr;
  input  logic [WORD_BITS               -1:0] in_init_data;

  // Debug interface
  output integer                              active_threads;


  logic [REQUEST_BITS -1:0] requests [0:NUM_DCORES];
  logic [NUM_DCORES                       :0] req_valids;
  logic [NUM_DCORES                       :0] req_readys;

  assign requests[0] = inp_request;
  assign req_valids[0] = inp_valid;
  assign out_ready = req_readys[0];
  assign req_readys[NUM_DCORES] = 1'b0;

  logic [REQUEST_BITS -1:0] c_requests [0:NUM_DCORES-1];
  logic [NUM_DCORES                     -1:0] c_req_valids;
  logic [NUM_DCORES                     -1:0] c_req_readys;



  logic [RESULT_BITS -1:0] c_results [0:NUM_DCORES-1];
  logic [NUM_DCORES                     -1:0] c_res_valids;
  logic [NUM_DCORES                     -1:0] c_res_readys;


  logic [RESULT_BITS -1:0] results [0:NUM_DCORES];
  logic [NUM_DCORES                       :0] res_valids;
  logic [NUM_DCORES                       :0] res_readys;

  assign out_result = results[NUM_DCORES];
  assign out_valid = res_valids[NUM_DCORES];
  assign res_readys[NUM_DCORES] = inp_ready;
  assign res_valids[0] = 1'b0;

  logic [NUM_DCORES                       :0] init_loads;
  logic [OBJ_BUFFER_ADDR_BITS -1:0] init_addrs [0:NUM_DCORES];
  logic [WORD_BITS -1:0] init_datas [0:NUM_DCORES];

  assign init_loads[0] = in_init_load;
  assign init_addrs[0] = in_init_addr;
  assign init_datas[0] = in_init_data;

  // Debug
  int j;
  integer active_threads_arr [0:NUM_DCORES-1];
  always_comb begin
    active_threads = 0;
    for (j = 0; j < NUM_DCORES; j++) begin
      active_threads += active_threads_arr[j];
    end
  end

  genvar i;
  for (i = 0; i < NUM_DCORES; i++) begin : gen_loop
    // if ( (i%8) == 7 & (i != (NUM_DCORES-1) )) begin
    if (0) begin
      feeder_full FEEDER_F (
        .reset_n(reset_n),
        .clk(clk),
        .slave_valid(req_valids[i]),
        .slave_ready(req_readys[i]),
        .slave_data(requests[i]),
        .master_valid(req_valids[i+1]),
        .master_ready(req_readys[i+1]),
        .master_data(requests[i+1]),
        .core_valid(c_req_valids[i]),
        .core_ready(c_req_readys[i]),
        .core_data(c_requests[i])
      );
    end
    else begin
      feeder FEEDER (
        .reset_n(reset_n),
        .clk(clk),
        .slave_valid(req_valids[i]),
        .slave_ready(req_readys[i]),
        .slave_data(requests[i]),
        .master_valid(req_valids[i+1]),
        .master_ready(req_readys[i+1]),
        .master_data(requests[i+1]),
        .core_valid(c_req_valids[i]),
        .core_ready(c_req_readys[i]),
        .core_data(c_requests[i])
      );
    end

    dual_core DCORE (
      .reset_n(reset_n),
      .clk(clk),
      .inp_request(c_requests[i]),
      .inp_valid(c_req_valids[i]),
      .out_ready(c_req_readys[i]),
      .out_result(c_results[i]),
      .out_valid(c_res_valids[i]),
      .inp_ready(c_res_readys[i]),
      .in_init_load(init_loads[i]),
      .in_init_addr(init_addrs[i]),
      .in_init_data(init_datas[i]),
      .out_init_load(init_loads[i+1]),
      .out_init_addr(init_addrs[i+1]),
      .out_init_data(init_datas[i+1]),
      .active_threads(active_threads_arr[i])
    );

    // if ( (i%8) == 7 & (i != (NUM_DCORES-1) )) begin
    if (0) begin
      drain_full DRAIN_F (
        .reset_n(reset_n),
        .clk(clk),
        .slave_valid(res_valids[i]),
        .slave_ready(res_readys[i]),
        .slave_data(results[i]),
        .master_valid(res_valids[i+1]),
        .master_ready(res_readys[i+1]),
        .master_data(results[i+1]),
        .core_valid(c_res_valids[i]),
        .core_ready(c_res_readys[i]),
        .core_data(c_results[i])
      );
    end
    else begin
      drain DRAIN (
        .reset_n(reset_n),
        .clk(clk),
        .slave_valid(res_valids[i]),
        .slave_ready(res_readys[i]),
        .slave_data(results[i]),
        .master_valid(res_valids[i+1]),
        .master_ready(res_readys[i+1]),
        .master_data(results[i+1]),
        .core_valid(c_res_valids[i]),
        .core_ready(c_res_readys[i]),
        .core_data(c_results[i])
      );
    end
  end


endmodule