/**
 * Module that adds/subtracts two IEEE754-2008 binary32 floating-point numbers.
 *
 * Nan/inf special cases are not supported
 * Rounding mode: roundTiesToEven
 * No support for sub-normal numbers
 */
module fp32_add_2s
  (
  // Outputs
  out_res,
  // Inputs
  clk, inp_a, inp_b, inp_op
  );

  parameter ROUND = 1;

  localparam FLOAT32_BITS     = 32;
  localparam FLOAT32_SGN_BITS = 1;
  localparam FLOAT32_EXP_BITS = 8;
  localparam FLOAT32_TSF_BITS = FLOAT32_BITS - FLOAT32_SGN_BITS - FLOAT32_EXP_BITS;
  localparam FLOAT32_SFC_BITS = FLOAT32_TSF_BITS + 1;
  localparam PRODUCT_BITS     = FLOAT32_SFC_BITS + FLOAT32_SFC_BITS;

  // Control
  input  logic                               clk;

  // Inputs
  input  logic [FLOAT32_BITS           -1:0] inp_a;
  input  logic [FLOAT32_BITS           -1:0] inp_b;
  input  logic                               inp_op;

  // Outputs
  output logic [FLOAT32_BITS           -1:0] out_res;


  //*****************************************************************************
  //
  // Pipeline step 0
  //
  // - Splitting inputs into sign, exponent and encoded significand.
  // - Swapping and aligning operands.
  // - Choosing larger exponent.
  // - Calculating effective operation.
  // - Complementing smaller significand if effective operation is sub.
  //   - enables use of 2's comp. adder.
  //*****************************************************************************
  logic [FLOAT32_SGN_BITS              -1:0] s0_a_sgn;
  logic [FLOAT32_EXP_BITS              -1:0] s0_a_exp;
  logic [FLOAT32_TSF_BITS              -1:0] s0_a_sfc_enc;
  logic [FLOAT32_SFC_BITS              -1:0] s0_a_sfc;

  logic [FLOAT32_SGN_BITS              -1:0] s0_b_sgn;
  logic [FLOAT32_EXP_BITS              -1:0] s0_b_exp;
  logic [FLOAT32_TSF_BITS              -1:0] s0_b_sfc_enc;
  logic [FLOAT32_SFC_BITS              -1:0] s0_b_sfc;

  logic                                      s0_a_nrm;
  logic                                      s0_b_nrm;

  // Splitting inputs into sign, exponent and encoded significand
  assign { s0_a_sgn, s0_a_exp, s0_a_sfc_enc } = inp_a;
  assign { s0_b_sgn, s0_b_exp, s0_b_sfc_enc } = inp_b;

  // Are the inputs normal numbers?
  assign s0_a_nrm = |s0_a_exp;
  assign s0_b_nrm = |s0_b_exp;

  // Assuming normal inputs - prepending hidden 1
  assign s0_a_sfc = { s0_a_nrm, s0_a_sfc_enc };
  assign s0_b_sfc = { s0_b_nrm, s0_b_sfc_enc };

  // Calculating exponent difference as well as its sign and absolute value
  logic [FLOAT32_EXP_BITS+1            -1:0] s0_d;
  logic                                      s0_d_sgn;
  logic [FLOAT32_EXP_BITS              -1:0] s0_d_abs;
  logic                                      s0_d_zero;
  assign s0_d = s0_a_exp - s0_b_exp;
  assign s0_d_sgn = s0_d[FLOAT32_EXP_BITS];
  assign s0_d_abs = ( s0_d_sgn ? (~s0_d)+1 : s0_d );
  assign s0_d_zero = ( s0_d == 0 );

  // Choosing larger exponent and swapping significands
  logic [FLOAT32_EXP_BITS              -1:0] s0_res_exp;
  assign s0_res_exp = ( s0_d_sgn ? s0_b_exp : s0_a_exp );
  logic [FLOAT32_SFC_BITS              -1:0] s0_x_sfc;
  logic [FLOAT32_SFC_BITS              -1:0] s0_y_sfc;
  assign s0_x_sfc = ( s0_d_sgn ? s0_b_sfc : s0_a_sfc );
  assign s0_y_sfc = ( s0_d_sgn ? s0_a_sfc : s0_b_sfc );

  // Aligning significands and calculating guard, round and sticky bit
  logic [FLOAT32_SFC_BITS+3            -1:0] s0_x_sfc_aligned;
  logic [FLOAT32_SFC_BITS+3            -1:0] s0_y_sfc_aligned;
  assign s0_x_sfc_aligned = { s0_x_sfc, 3'b000 };
  always_comb begin
    unique case ( s0_d_abs )
      // Sticky bit           -----------------------------------------------------------------\
      // Guard and round bits ----------------------------------------\                        |
      // Aligned significand  ------------\                           |                        |
      //                                  |                           |                        |
      //                        ______________________    ___________________________    _______________
        0: s0_y_sfc_aligned = {        s0_y_sfc[23: 0],         2'b00,                   1'b0            };
        1: s0_y_sfc_aligned = {  1'd0, s0_y_sfc[23: 1],         s0_y_sfc[  0  ], 1'b0,   1'b0            };
        2: s0_y_sfc_aligned = {  2'd0, s0_y_sfc[23: 2],         s0_y_sfc[ 1: 0],         1'b0            };
        3: s0_y_sfc_aligned = {  3'd0, s0_y_sfc[23: 3],         s0_y_sfc[ 2: 1],         |s0_y_sfc[ 0:0] };
        4: s0_y_sfc_aligned = {  4'd0, s0_y_sfc[23: 4],         s0_y_sfc[ 3: 2],         |s0_y_sfc[ 1:0] };
        5: s0_y_sfc_aligned = {  5'd0, s0_y_sfc[23: 5],         s0_y_sfc[ 4: 3],         |s0_y_sfc[ 2:0] };
        6: s0_y_sfc_aligned = {  6'd0, s0_y_sfc[23: 6],         s0_y_sfc[ 5: 4],         |s0_y_sfc[ 3:0] };
        7: s0_y_sfc_aligned = {  7'd0, s0_y_sfc[23: 7],         s0_y_sfc[ 6: 5],         |s0_y_sfc[ 4:0] };
        8: s0_y_sfc_aligned = {  8'd0, s0_y_sfc[23: 8],         s0_y_sfc[ 7: 6],         |s0_y_sfc[ 5:0] };
        9: s0_y_sfc_aligned = {  9'd0, s0_y_sfc[23: 9],         s0_y_sfc[ 8: 7],         |s0_y_sfc[ 6:0] };
       10: s0_y_sfc_aligned = { 10'd0, s0_y_sfc[23:10],         s0_y_sfc[ 9: 8],         |s0_y_sfc[ 7:0] };
       11: s0_y_sfc_aligned = { 11'd0, s0_y_sfc[23:11],         s0_y_sfc[10: 9],         |s0_y_sfc[ 8:0] };
       12: s0_y_sfc_aligned = { 12'd0, s0_y_sfc[23:12],         s0_y_sfc[11:10],         |s0_y_sfc[ 9:0] };
       13: s0_y_sfc_aligned = { 13'd0, s0_y_sfc[23:13],         s0_y_sfc[12:11],         |s0_y_sfc[10:0] };
       14: s0_y_sfc_aligned = { 14'd0, s0_y_sfc[23:14],         s0_y_sfc[13:12],         |s0_y_sfc[11:0] };
       15: s0_y_sfc_aligned = { 15'd0, s0_y_sfc[23:15],         s0_y_sfc[14:13],         |s0_y_sfc[12:0] };
       16: s0_y_sfc_aligned = { 16'd0, s0_y_sfc[23:16],         s0_y_sfc[15:14],         |s0_y_sfc[13:0] };
       17: s0_y_sfc_aligned = { 17'd0, s0_y_sfc[23:17],         s0_y_sfc[16:15],         |s0_y_sfc[14:0] };
       18: s0_y_sfc_aligned = { 18'd0, s0_y_sfc[23:18],         s0_y_sfc[17:16],         |s0_y_sfc[15:0] };
       19: s0_y_sfc_aligned = { 19'd0, s0_y_sfc[23:19],         s0_y_sfc[18:17],         |s0_y_sfc[16:0] };
       20: s0_y_sfc_aligned = { 20'd0, s0_y_sfc[23:20],         s0_y_sfc[19:18],         |s0_y_sfc[17:0] };
       21: s0_y_sfc_aligned = { 21'd0, s0_y_sfc[23:21],         s0_y_sfc[20:19],         |s0_y_sfc[18:0] };
       22: s0_y_sfc_aligned = { 22'd0, s0_y_sfc[23:22],         s0_y_sfc[21:20],         |s0_y_sfc[19:0] };
       23: s0_y_sfc_aligned = { 23'd0, s0_y_sfc[23:23],         s0_y_sfc[22:21],         |s0_y_sfc[20:0] };
       24: s0_y_sfc_aligned = { 24'd0,                          s0_y_sfc[23:22],         |s0_y_sfc[21:0] };
       25: s0_y_sfc_aligned = { 24'd0,                    1'b0, s0_y_sfc[  23 ],         |s0_y_sfc[22:0] };
       26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41,
       42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57,
       58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73,
       74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89,
       90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104,
      105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117,
      118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130,
      131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143,
      144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156,
      157, 158, 159, 160, 161, 162, 163, 164, 165, 166, 167, 168, 169,
      170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182,
      183, 184, 185, 186, 187, 188, 189, 190, 191, 192, 193, 194, 195,
      196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208,
      209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221,
      222, 223, 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234,
      235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247,
      248, 249, 250, 251, 252, 253, 254,
      255: s0_y_sfc_aligned = { 24'd0,                          2'b00,                   |s0_y_sfc[23:0] };

      default: s0_y_sfc_aligned = { FLOAT32_SFC_BITS+3 { 1'bx } };
    endcase
  end

  // Calculating effective operation
  logic                                      s0_eop;
  assign s0_eop = (s0_a_sgn ^ s0_b_sgn) ^ inp_op;

  // Complementing smaller significand in case of effective subtraction
  logic                                      s0_a_sfc_lt_b_sfc;
  logic                                      s0_comp_x;
  logic                                      s0_comp_y;
  logic [FLOAT32_SFC_BITS+3+1          -1:0] s0_x_sfc_aligned_comp;
  logic [FLOAT32_SFC_BITS+3+1          -1:0] s0_y_sfc_aligned_comp;
  logic [FLOAT32_SFC_BITS+3+1          -1:0] s0_x_sfc_aligned_comp_r;
  logic [FLOAT32_SFC_BITS+3+1          -1:0] s0_y_sfc_aligned_comp_r;
  assign s0_a_sfc_lt_b_sfc = s0_a_sfc_enc < s0_b_sfc_enc;
  assign s0_comp_x = s0_eop & (s0_d_abs == 0) & s0_a_sfc_lt_b_sfc;
  assign s0_comp_y = s0_eop & !s0_comp_x;

  assign s0_x_sfc_aligned_comp = ( s0_comp_x ? { 1'b1, ~s0_x_sfc_aligned } : { 1'b0, s0_x_sfc_aligned });
  assign s0_y_sfc_aligned_comp = ( s0_comp_y ? { 1'b1, ~s0_y_sfc_aligned } : { 1'b0, s0_y_sfc_aligned });


  //** Register results of pipeline stage 0
  logic [FLOAT32_SGN_BITS              -1:0] s0_a_sgn_r;
  logic [FLOAT32_SGN_BITS              -1:0] s0_b_sgn_r;
  logic                                      s0_d_sgn_r;
  logic                                      s0_d_zero_r;
  logic                                      s0_eop_r;
  logic                                      s0_a_sfc_lt_b_sfc_r;
  logic [FLOAT32_EXP_BITS              -1:0] s0_res_exp_r;
  always_ff @( posedge clk ) begin
    s0_a_sgn_r <= s0_a_sgn;
    s0_b_sgn_r <= s0_b_sgn;
    s0_d_sgn_r <= s0_d_sgn;
    s0_d_zero_r <= s0_d_zero;
    s0_eop_r <= s0_eop;
    s0_a_sfc_lt_b_sfc_r <= s0_a_sfc_lt_b_sfc;
    s0_res_exp_r <= s0_res_exp;
    s0_x_sfc_aligned_comp_r <= s0_x_sfc_aligned_comp;
    s0_y_sfc_aligned_comp_r <= s0_y_sfc_aligned_comp;
  end


  //*****************************************************************************
  //
  // Pipeline step 1
  // - Performing actual addition.
  //*****************************************************************************


  // FINALLY doing the addition
  logic [FLOAT32_SFC_BITS+3+1          -1:0] s1_sum;
  logic [FLOAT32_SFC_BITS+3+1          -1:0] s1_sum_r;
  assign s1_sum = s0_x_sfc_aligned_comp_r + s0_y_sfc_aligned_comp_r + s0_eop_r;


  logic [FLOAT32_SGN_BITS              -1:0] s1_a_sgn_r;
  logic [FLOAT32_SGN_BITS              -1:0] s1_b_sgn_r;
  logic                                      s1_d_sgn_r;
  logic                                      s1_d_zero_r;
  logic                                      s1_eop_r;
  logic                                      s1_a_sfc_lt_b_sfc_r;
  logic [FLOAT32_EXP_BITS              -1:0] s1_res_exp_r;
  always_ff @( posedge clk ) begin
    s1_sum_r <= s1_sum;
    s1_a_sgn_r <= s0_a_sgn_r;
    s1_b_sgn_r <= s0_b_sgn_r;
    s1_d_sgn_r <= s0_d_sgn_r;
    s1_d_zero_r <= s0_d_zero_r;
    s1_eop_r <= s0_eop_r;
    s1_a_sfc_lt_b_sfc_r <= s0_a_sfc_lt_b_sfc_r;
    s1_res_exp_r <= s0_res_exp_r;
  end


  //*****************************************************************************
  //
  // Pipeline step 2
  //
  // - Aligning sum and calculating number of leading zeroes.
  // - Performing rounding.
  // - Realigning significand in case of overflow after rounding.
  // - Calculating final exponent by subtracting # leading zeroes
  //   and correcting for overflow.
  //   - This could probably be done in the same way as for mult.
  //     - And the logic can thus be shared as well.
  //*****************************************************************************

  // Aligning sum and calculating number of leading zeroes
  logic [FLOAT32_SFC_BITS+3            -1:0] s1_sum_aligned;
  logic [FLOAT32_EXP_BITS              -1:0] s1_sum_lz;
  always_comb begin
    unique casez ( s1_sum )
      // Leading zeroes          ---------------------------------------------------------------------------\
      // Guard bits and aligning -----------------------------------------------------------------\         |
      // Aligned significand     ----------------------------------------------\                  |         |
      //                                                                       |                  |         |
      28'b1?_??????????????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[27:2], |s1_sum[1:0], 8'd255}; // -1
      28'b01_??????????????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[26:1], |s1_sum[0:0], 8'd0  };
      28'b00_1?????????????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[25:0],           1'd0, 8'd1  };
      28'b00_01????????????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[24:0],           2'd0, 8'd2  };
      28'b00_001???????????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[23:0],           3'd0, 8'd3  };
      28'b00_0001??????????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[22:0],           4'd0, 8'd4  };
      28'b00_00001?????????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[21:0],           5'd0, 8'd5  };
      28'b00_000001????????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[20:0],           6'd0, 8'd6  };
      28'b00_0000001???????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[19:0],           7'd0, 8'd7  };
      28'b00_00000001??????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[18:0],           8'd0, 8'd8  };
      28'b00_000000001?????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[17:0],           9'd0, 8'd9  };
      28'b00_0000000001????????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[16:0],          10'd0, 8'd10 };
      28'b00_00000000001???????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[15:0],          11'd0, 8'd11 };
      28'b00_000000000001??????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[14:0],          12'd0, 8'd12 };
      28'b00_0000000000001?????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[13:0],          13'd0, 8'd13 };
      28'b00_00000000000001????????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[12:0],          14'd0, 8'd14 };
      28'b00_000000000000001???????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[11:0],          15'd0, 8'd15 };
      28'b00_0000000000000001??????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[10:0],          16'd0, 8'd16 };
      28'b00_00000000000000001?????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[ 9:0],          17'd0, 8'd17 };
      28'b00_000000000000000001????????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[ 8:0],          18'd0, 8'd18 };
      28'b00_0000000000000000001???????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[ 7:0],          19'd0, 8'd19 };
      28'b00_00000000000000000001??????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[ 6:0],          20'd0, 8'd20 };
      28'b00_000000000000000000001?????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[ 5:0],          21'd0, 8'd21 };
      28'b00_0000000000000000000001????: { s1_sum_aligned, s1_sum_lz } = { s1_sum[ 4:0],          22'd0, 8'd22 };
      28'b00_00000000000000000000001???: { s1_sum_aligned, s1_sum_lz } = { s1_sum[ 3:0],          23'd0, 8'd23 };
      28'b00_00000000000000000000000???: { s1_sum_aligned, s1_sum_lz } = { 3'd0,                    24'd0, 8'd24 };
      
      default                          : { s1_sum_aligned, s1_sum_lz } = { FLOAT32_SFC_BITS+3+FLOAT32_EXP_BITS { 1'bx } };
    endcase
  end

  logic [FLOAT32_SFC_BITS+3            -1:0] s1_sum_aligned_r;
  logic [FLOAT32_EXP_BITS              -1:0] s1_sum_lz_r;

  always_ff @(posedge clk) begin
    s1_sum_aligned_r <= s1_sum_aligned;
    s1_sum_lz_r <= s1_sum_lz;
  end

  // Splitting aligned result into significand, guard, round and sticky bit
  logic [FLOAT32_SFC_BITS              -1:0] s2_sum_aligned_noguard;
  logic                                      s2_guard;
  logic                                      s2_round;
  logic                                      s2_sticky;
  assign { s2_sum_aligned_noguard, s2_guard, s2_round, s2_sticky } = s1_sum_aligned_r;

  // Should we round up?
  logic                                      s2_round_up;
  generate
    if (ROUND == 1) begin
      assign s2_round_up = s2_guard & ( s2_sum_aligned_noguard[0] | s2_round | s2_sticky);
    end
    else begin
      assign s2_round_up = 0;
    end
  endgenerate

  // Recalculating exponent
  logic [FLOAT32_EXP_BITS+1            -1:0] s2_res_exp_aligned;
  assign s2_res_exp_aligned = s1_res_exp_r - { s1_sum_lz_r[7], s1_sum_lz_r };

  // Is the significand zero?
  logic                                      s2_sfc_zero;
  assign s2_sfc_zero = ( s2_sum_aligned_noguard == 0 );

  // Did the exponent underflow?
  logic                                      s2_res_exp_unf;
  assign s2_res_exp_unf = s2_res_exp_aligned[FLOAT32_EXP_BITS] | ~|s2_res_exp_aligned[FLOAT32_EXP_BITS-1:0];

  logic [FLOAT32_BITS-1                -1:0] s2_res_concat;
  always_comb begin
    unique if (s2_res_exp_unf | s2_sfc_zero) begin
      s2_res_concat = { { FLOAT32_EXP_BITS { 1'b0 } }, { FLOAT32_TSF_BITS {1'b0} } };
    end
    else begin
      s2_res_concat = { s2_res_exp_aligned, s2_sum_aligned_noguard[FLOAT32_TSF_BITS-1:0] } + s2_round_up;
    end
  end


  // Sign calculation
  logic [6                               :0] s2_sgn_input;
  logic                                      s2_final_sgn;
  assign s1_sgn_input = { s1_d_sgn_r, s1_d_zero_r, s1_a_sgn_r, s1_b_sgn_r, s1_eop_r, s1_a_sfc_lt_b_sfc_r, s2_sfc_zero };
  always_comb begin
    unique case ( s1_sgn_input )
      7'b0000000: s2_final_sgn = 0;
      7'b0000001: s2_final_sgn = 0;
      7'b0000010: s2_final_sgn = 0;
      7'b0000100: s2_final_sgn = 0;
      7'b0000101: s2_final_sgn = 0;
      7'b0000110: s2_final_sgn = 0;
      7'b0001000: s2_final_sgn = 0;
      7'b0001001: s2_final_sgn = 0;
      7'b0001010: s2_final_sgn = 0;
      7'b0001100: s2_final_sgn = 0;
      7'b0001101: s2_final_sgn = 0;
      7'b0001110: s2_final_sgn = 0;
      7'b0010000: s2_final_sgn = 1;
      7'b0010001: s2_final_sgn = 1;
      7'b0010010: s2_final_sgn = 1;
      7'b0010100: s2_final_sgn = 1;
      7'b0010101: s2_final_sgn = 1;
      7'b0010110: s2_final_sgn = 1;
      7'b0011000: s2_final_sgn = 1;
      7'b0011001: s2_final_sgn = 1;
      7'b0011010: s2_final_sgn = 1;
      7'b0011100: s2_final_sgn = 1;
      7'b0011101: s2_final_sgn = 1;
      7'b0011110: s2_final_sgn = 1;
      7'b0100000: s2_final_sgn = 0;
      7'b0100001: s2_final_sgn = 0;
      7'b0100010: s2_final_sgn = 0;
      7'b0100100: s2_final_sgn = 0;
      7'b0100101: s2_final_sgn = 0;
      7'b0100110: s2_final_sgn = 1;
      7'b0101000: s2_final_sgn = 0;
      7'b0101001: s2_final_sgn = 0;
      7'b0101010: s2_final_sgn = 0;
      7'b0101100: s2_final_sgn = 0;
      7'b0101101: s2_final_sgn = 0;
      7'b0101110: s2_final_sgn = 1;
      7'b0110000: s2_final_sgn = 1;
      7'b0110001: s2_final_sgn = 1;
      7'b0110010: s2_final_sgn = 1;
      7'b0110100: s2_final_sgn = 1;
      7'b0110101: s2_final_sgn = 0;
      7'b0110110: s2_final_sgn = 0;
      7'b0111000: s2_final_sgn = 1;
      7'b0111001: s2_final_sgn = 1;
      7'b0111010: s2_final_sgn = 1;
      7'b0111100: s2_final_sgn = 1;
      7'b0111101: s2_final_sgn = 0;
      7'b0111110: s2_final_sgn = 0;
      7'b1000000: s2_final_sgn = 0;
      7'b1000001: s2_final_sgn = 0;
      7'b1000010: s2_final_sgn = 0;
      7'b1000100: s2_final_sgn = 1;
      7'b1000101: s2_final_sgn = 1;
      7'b1000110: s2_final_sgn = 1;
      7'b1001000: s2_final_sgn = 0;
      7'b1001001: s2_final_sgn = 0;
      7'b1001010: s2_final_sgn = 0;
      7'b1001100: s2_final_sgn = 1;
      7'b1001101: s2_final_sgn = 1;
      7'b1001110: s2_final_sgn = 1;
      7'b1010000: s2_final_sgn = 1;
      7'b1010001: s2_final_sgn = 1;
      7'b1010010: s2_final_sgn = 1;
      7'b1010100: s2_final_sgn = 0;
      7'b1010101: s2_final_sgn = 0;
      7'b1010110: s2_final_sgn = 0;
      7'b1011000: s2_final_sgn = 1;
      7'b1011001: s2_final_sgn = 1;
      7'b1011010: s2_final_sgn = 1;
      7'b1011100: s2_final_sgn = 0;
      7'b1011101: s2_final_sgn = 0;
      7'b1011110: s2_final_sgn = 0;
      
      default   : s2_final_sgn = 1'bx;
    endcase
  end

  assign out_res = { s2_final_sgn, s2_res_concat };

endmodule