/**
 * Module that multiplies two IEEE754-2008 binary32 floating-point numbers.
 *
 * Nan/inf special cases are not supported
 * Rounding mode: roundTiesToEven
 * No support for sub-normal numbers
 */
module fp32_mul_2s
  (
  // Outputs
  out_res,
  // Inputs
  clk, inp_a, inp_b
  );

  parameter ROUND = 1;

  localparam FLOAT32_BITS     = 32;
  localparam FLOAT32_SGN_BITS = 1;
  localparam FLOAT32_EXP_BITS = 8;
  localparam FLOAT32_TSF_BITS = FLOAT32_BITS - FLOAT32_SGN_BITS - FLOAT32_EXP_BITS;
  localparam FLOAT32_SFC_BITS = FLOAT32_TSF_BITS + 1;
  localparam PRODUCT_BITS     = FLOAT32_SFC_BITS + FLOAT32_SFC_BITS;

  // Control
  input  logic                               clk;

  // Inputs
  input  logic [FLOAT32_BITS           -1:0] inp_a;
  input  logic [FLOAT32_BITS           -1:0] inp_b;

  // Outputs
  output logic [FLOAT32_BITS           -1:0] out_res;


  //*****************************************************************************
  //
  // Pipeline step 0
  //
  // - Splitting inputs into sign, exponent and encoded significand.
  // - Multiplying full significands
  // - Adding exponents
  //*****************************************************************************
  logic [FLOAT32_SGN_BITS              -1:0] s0_a_sgn;
  logic [FLOAT32_EXP_BITS              -1:0] s0_a_exp;
  logic [FLOAT32_TSF_BITS              -1:0] s0_a_sfc_enc;
  logic [FLOAT32_SFC_BITS              -1:0] s0_a_sfc;

  logic [FLOAT32_SGN_BITS              -1:0] s0_b_sgn;
  logic [FLOAT32_EXP_BITS              -1:0] s0_b_exp;
  logic [FLOAT32_TSF_BITS              -1:0] s0_b_sfc_enc;
  logic [FLOAT32_SFC_BITS              -1:0] s0_b_sfc;

  // Splitting inputs into sign, exponent and encoded significand
  assign { s0_a_sgn, s0_a_exp, s0_a_sfc_enc } = inp_a;
  assign { s0_b_sgn, s0_b_exp, s0_b_sfc_enc } = inp_b;

  // Assuming normal inputs - prepending hidden 1
  assign s0_a_sfc = { 1'b1, s0_a_sfc_enc };
  assign s0_b_sfc = { 1'b1, s0_b_sfc_enc };

  // Multiplying significands
  logic [PRODUCT_BITS                  -1:0] s0_prd_sfc_nxt;
  assign s0_prd_sfc_nxt = s0_a_sfc * s0_b_sfc;

  // Adding exponents
  logic [FLOAT32_EXP_BITS+2            -1:0] s0_prd_exp_nxt;
  assign s0_prd_exp_nxt = { 2'b00, s0_a_exp } + { 2'b00, s0_b_exp } - 10'd127;

  // Computing sign of result
  logic [FLOAT32_SGN_BITS              -1:0] s0_res_sgn_nxt;
  assign s0_res_sgn_nxt = s0_a_sgn ^ s0_b_sgn;
  

  //** Double register results of pipeline stage 0
  logic [PRODUCT_BITS                  -1:0] s00_prd_sfc_r;

  logic [FLOAT32_EXP_BITS+2            -1:0] s00_prd_exp_r;

  logic [FLOAT32_SGN_BITS              -1:0] s00_res_sgn_r;
  
  logic [FLOAT32_EXP_BITS              -1:0] s00_a_exp_r;
  logic [FLOAT32_EXP_BITS              -1:0] s00_b_exp_r;

  logic [PRODUCT_BITS                  -1:0] s0_prd_sfc_r;

  logic [FLOAT32_EXP_BITS+2            -1:0] s0_prd_exp_r;

  logic [FLOAT32_SGN_BITS              -1:0] s0_res_sgn_r;
  
  logic [FLOAT32_EXP_BITS              -1:0] s0_a_exp_r;
  logic [FLOAT32_EXP_BITS              -1:0] s0_b_exp_r;

  always_ff @(posedge clk) begin
    s0_prd_sfc_r <= s00_prd_sfc_r;

    s0_prd_exp_r <= s00_prd_exp_r;

    s0_res_sgn_r <= s00_res_sgn_r;

    s0_a_exp_r <= s00_a_exp_r;
    s0_b_exp_r <= s00_b_exp_r;

    s00_prd_sfc_r <= s0_prd_sfc_nxt;

    s00_prd_exp_r <= s0_prd_exp_nxt;

    s00_res_sgn_r <= s0_res_sgn_nxt;

    s00_a_exp_r <= s0_a_exp;
    s00_b_exp_r <= s0_b_exp;
  end


  //*****************************************************************************
  //
  // Pipeline step 1
  //
  // Normalising result if necessary
  // Rounding result
  // Handling special cases such as zero/denormal operands
  //*****************************************************************************

  // Does the resulting significand require normalisation?
  logic                                      s1_prd_sfc_ovf;
  assign s1_prd_sfc_ovf = s0_prd_sfc_r[47];

  // Normalise significand if needed
  logic [FLOAT32_TSF_BITS              -1:0] s1_res_sfc;
  assign s1_res_sfc = ( s1_prd_sfc_ovf
                 ? s0_prd_sfc_r[2*FLOAT32_TSF_BITS+1 -1:FLOAT32_TSF_BITS+1]
                 : s0_prd_sfc_r[2*FLOAT32_TSF_BITS -1:FLOAT32_TSF_BITS] );

  // Increment exponent if normalisation was done
  logic [FLOAT32_EXP_BITS+2            -1:0] s1_prd_exp_norm;
  assign s1_prd_exp_norm = s0_prd_exp_r + s1_prd_sfc_ovf;

  // Get round bit from significand
  logic                                      s1_prd_sfc_rnd;
  assign s1_prd_sfc_rnd = ( s1_prd_sfc_ovf
                     ? s0_prd_sfc_r[23]
                     : s0_prd_sfc_r[22] );

  // Compute sticky bit from significand
  logic                                      s1_prd_sfc_sty;
  assign s1_prd_sfc_sty = |{(s1_prd_sfc_ovf & s0_prd_sfc_r[22]), s0_prd_sfc_r[21:0]};

  // Should we round up?
  logic                                      s1_round_up;
  generate
    if (ROUND == 1) begin
      assign s1_round_up = s1_prd_sfc_rnd & (s1_prd_sfc_sty | s1_res_sfc[0]);
    end
    else begin
      assign s1_round_up = 0;
    end
  endgenerate

  // Did the exponent underflow?
  logic                                      s1_prd_exp_unf;
  assign s1_prd_exp_unf = s1_prd_exp_norm[FLOAT32_EXP_BITS+1] | ~|s1_prd_exp_norm[FLOAT32_EXP_BITS:0];

  // Flush-to-zero if underflow. Set result to zero if subnormal inputs.
  // Or apply rounding.
  logic [FLOAT32_BITS-1                -1:0] s1_sfc_res_rnd;
  always_comb begin
    unique if (s1_prd_exp_unf | (~|s0_a_exp_r) | (~|s0_b_exp_r)) begin
      s1_sfc_res_rnd = { { FLOAT32_EXP_BITS { 1'b0 } }, { FLOAT32_TSF_BITS {1'b0} } };
    end
    else begin
      s1_sfc_res_rnd = { s1_prd_exp_norm, s1_res_sfc } + s1_round_up;
    end
  end

  logic [FLOAT32_BITS                  -1:0] out_res_nxt;
  assign out_res = {s0_res_sgn_r, s1_sfc_res_rnd};

endmodule