/**
 * Interface between whole raytracer system and an individual core.
 */
module ray_core
  (
    // Outputs
    out_ready, out_result, out_valid, out_obj_read, out_obj_addr, inp_instruction,
    // Inputs
    clk, reset_n, inp_request, inp_valid, inp_ready, inp_obj_data, out_instruction_address,
    // Debug
    active_threads
  );

  import types_pkg::*;


  // Control
  input  logic                                reset_n;
  input  logic                                clk;

  // Request interface
  input  logic [REQUEST_BITS            -1:0] inp_request;
  input  logic                                inp_valid;
  output logic                                out_ready;

  // Result interface
  output logic [RESULT_BITS             -1:0] out_result;
  output logic                                out_valid;
  input  logic                                inp_ready;

  // Object buffer interface
  output logic [OBJ_BUFFER_ADDR_BITS    -1:0] out_obj_addr;
  output logic                                out_obj_read;
  input  logic [WORD_BITS               -1:0] inp_obj_data;

  // Instruction memory interface
  output logic [PC_BITS                 -1:0] out_instruction_address;
  input  instruction_union_t                  inp_instruction;

  // Debug interface
  output integer                              active_threads;


  // Core signals
  logic [WORD_BITS               -1:0] c_inp_req;
  command_t                            c_inp_command;

  logic                                c_out_req_shift;
  logic [BYTE_BITS               -1:0] c_out_result;
  logic                                c_out_result_valid;
  logic [$clog2(THREADS)         -1:0] c_out_thread;
  status_t                             c_out_status;
  logic                                c_out_status_valid;


  // core_fpu_2s CORE (
  ray_datapath CORE (
    .reset_n(reset_n),
    .clk(clk),

    .inp_req(c_inp_req),
    .inp_command(c_inp_command),
    .inp_obj_data(inp_obj_data),
    .inp_instruction(inp_instruction),

    .out_req_shift(c_out_req_shift),
    .out_result(c_out_result),
    .out_result_valid(c_out_result_valid),
    .out_thread(c_out_thread),
    .out_status(c_out_status),
    .out_status_valid(c_out_status_valid),
    .out_obj_addr(out_obj_addr),
    .out_obj_read(out_obj_read),
    .out_instruction_address(out_instruction_address)
  );


  //*****************************************************************************
  //
  // Request handler FSM
  //
  // Controls incoming requests and assigns them to a thread.
  // Waits for core to read in request before accepting new requests.
  // Stores status output from core.
  //*****************************************************************************

  typedef enum logic[1:0] {
    IDLE_S,
    ARBIT_S,
    WAIT_BUSY_S
  } req_handler_state_t;

  logic [REQUEST_BITS                   -1:0] request_r;

  logic [SCREEN_COORD_BITS              -1:0] screen_coord_r [0:THREADS -1];
  logic                                       frame_id_r     [0:THREADS -1];
  command_t                                   command_r      [0:THREADS -1];
  status_t                                    status_r       [0:THREADS -1];

  logic [$clog2(THREADS)                -1:0] thread_slot_select;
  logic [$clog2(THREADS)                -1:0] thread_slot_select_r;
  logic                                       thread_slot_select_valid;

  req_handler_state_t                         req_handler_state_r;
  req_handler_state_t                         req_handler_state_nxt;

  logic                                       request_accept;

  assign request_accept = inp_valid & out_ready;


  // Give core command according to thread currently in S2
  assign c_inp_command = command_r[c_out_thread];
  // Camera direction input to core.
  assign c_inp_req = request_r[WORD_BITS -1:0];

  // Registering input request and shifting it when core requests that
  always_ff @(posedge clk) begin
    if (request_accept) begin
      request_r <= inp_request;
    end
    else if (c_out_req_shift) begin
      request_r <= {32'd0, request_r[REQUEST_BITS -1:WORD_BITS]};
    end
  end

  // Setting status register when core wants to update the status
  always_ff @(posedge clk or negedge reset_n) begin
    if ( !reset_n ) begin
      status_r <= '{default:IDLE};
    end
    else if (c_out_status_valid) begin
      status_r[c_out_thread] <= c_out_status;
    end
  end

  // Request handler FSM state register
  always_ff @(posedge clk or negedge reset_n) begin
    if ( !reset_n ) begin
      req_handler_state_r <= IDLE_S;
    end
    else begin
      req_handler_state_r <= req_handler_state_nxt;
    end
  end


  logic                                       save_request_en;
  logic                                       req_handler_core_working;

  // Saves frame_id and screen_coord for given request
  // Also saves the arbitration winner for later
  always_ff @(posedge clk) begin
    if (save_request_en) begin
        thread_slot_select_r <= thread_slot_select;
        frame_id_r[thread_slot_select] <= request_r[REQUEST_BITS -1];
        screen_coord_r[thread_slot_select] <= request_r[REQUEST_BITS -2:3*WORD_BITS];
    end
  end

  // Next state logic for request handler FSM.
  always_comb begin
    req_handler_state_nxt = req_handler_state_r;
    out_ready = 0;
    save_request_en = 0;
    req_handler_core_working = 0;
    unique case ( req_handler_state_r )
      IDLE_S:
        begin
          if ( request_accept ) begin
            req_handler_state_nxt = ARBIT_S;
          end
          out_ready = 1;
        end
      ARBIT_S:
        if ( thread_slot_select_valid ) begin
          req_handler_state_nxt = WAIT_BUSY_S;
          save_request_en = 1;
        end
      WAIT_BUSY_S:
        if ( status_r[thread_slot_select_r] == BUSY ) begin
          req_handler_state_nxt = IDLE_S;
          req_handler_core_working = 1;
        end
    endcase
  end

  // Request handler thread slot select logic.
  always_comb begin
    thread_slot_select_valid = 1;
    thread_slot_select = 2'b00;

    if          ((status_r[0] == IDLE) & (command_r[0] == NONE)) begin
      thread_slot_select = 2'b00;
    end else if ((status_r[1] == IDLE) & (command_r[1] == NONE)) begin
      thread_slot_select = 2'b01;
    end else if ((status_r[2] == IDLE) & (command_r[2] == NONE)) begin
      thread_slot_select = 2'b10;
    end else if ((status_r[3] == IDLE) & (command_r[3] == NONE)) begin
      thread_slot_select = 2'b11;
    end else begin
      thread_slot_select_valid = 0;
    end
  end



  //*****************************************************************************
  //
  // Result handler FSM
  //
  // Grants access to result buffer and asserts the out_valid signal
  //*****************************************************************************

  typedef enum logic[2:0] {
    WAIT_S,
    WAIT_2_S,
    WAIT_3_S,
    WAIT_4_S,
    SEND_OUT_S,
    WAIT_THREAD
  } result_handler_state_t;

  result_handler_state_t                      result_handler_state_r;
  result_handler_state_t                      result_handler_state_nxt;

  logic [$clog2(THREADS)                -1:0] arbited;
  logic [$clog2(THREADS)                -1:0] arbited_r;
  logic                                       arbited_valid;

  logic [RESULT_BITS                    -1:0] out_result_r;

  logic                                       load_resbuf;
  logic                                       result_accept;

  assign result_accept = inp_ready & out_valid;

  // Result handler FSM state register
  always_ff @(posedge clk or negedge reset_n) begin
    if ( !reset_n ) begin
      result_handler_state_r <= WAIT_S;
    end
    else begin
      result_handler_state_r <= result_handler_state_nxt;
    end
  end

  // Result handler arbiter
  always_comb begin
    arbited_valid = 1;
    arbited = 2'b00;
    if          ((status_r[0] == RESBUF_REQ) & (command_r[0] == NONE)) begin
      arbited = 2'b00;
    end else if ((status_r[1] == RESBUF_REQ) & (command_r[1] == NONE)) begin
      arbited = 2'b01;
    end else if ((status_r[2] == RESBUF_REQ) & (command_r[2] == NONE)) begin
      arbited = 2'b10;
    end else if ((status_r[3] == RESBUF_REQ) & (command_r[3] == NONE)) begin
      arbited = 2'b11;
    end else begin
      arbited_valid = 0;
    end
  end

  // Result handler FSM next state logic
  always_comb begin
    result_handler_state_nxt = result_handler_state_r;
    out_valid = 0;
    load_resbuf = 0;
    unique case ( result_handler_state_r )
      WAIT_S:
        if ( arbited_valid ) begin
          result_handler_state_nxt = WAIT_2_S;
          load_resbuf = 1;
        end
      WAIT_2_S:
        if ( c_out_result_valid )
          result_handler_state_nxt = WAIT_3_S;
      WAIT_3_S:
        if ( c_out_result_valid )
          result_handler_state_nxt = WAIT_4_S;
      WAIT_4_S:
        if ( c_out_result_valid )
          result_handler_state_nxt = SEND_OUT_S;
      SEND_OUT_S:
        begin
          out_valid = 1;
          if ( result_accept )
            result_handler_state_nxt = WAIT_THREAD;
        end
      WAIT_THREAD:
        if ( status_r[arbited_r] == IDLE)
          result_handler_state_nxt = WAIT_S;

    endcase
  end

  // Saving arbitration winner for later.
  // Initialising out_result_r and shifting in results as arbitration winner pushes them.
  always_ff @(posedge clk) begin
    if (load_resbuf) begin
      arbited_r <= arbited;
      out_result_r <= {24'd0, frame_id_r[arbited], screen_coord_r[arbited]};
    end
    else if (c_out_result_valid) begin
      out_result_r <= {out_result_r[RESULT_BITS-BYTE_BITS -1:0], c_out_result};
    end
  end


  command_t                                   request_handler_command;
  logic [$clog2(THREADS)                -1:0] request_handler_command_thread;
  logic                                       request_handler_command_valid;

  command_t                                   response_handler_command;
  logic [$clog2(THREADS)                -1:0] response_handler_command_thread;
  logic                                       response_handler_command_valid;

  always_ff @(posedge clk or negedge reset_n) begin
    if ( !reset_n ) begin
      command_r <= '{default:NONE};
    end
    else begin
      assert ((request_handler_command_thread != response_handler_command_thread) | !(request_handler_command_valid & response_handler_command_valid))
        else $error("Both request handler and response handler attempts to alter command of same thread!");

      if (request_handler_command_valid) begin
        command_r[request_handler_command_thread] <= request_handler_command;
      end

      if (response_handler_command_valid) begin
        command_r[response_handler_command_thread] <= response_handler_command;
      end
    end
  end


  always_comb begin
    request_handler_command = NONE;
    request_handler_command_thread = 0;
    request_handler_command_valid = 0;

    unique if (save_request_en) begin
      request_handler_command = START;
      request_handler_command_thread = thread_slot_select;
      request_handler_command_valid = 1;
    end
    else if (req_handler_core_working) begin
      request_handler_command = NONE;
      request_handler_command_thread = thread_slot_select_r;
      request_handler_command_valid = 1;
    end
    else begin
      request_handler_command_valid = 0;
    end


    response_handler_command = NONE;
    response_handler_command_thread = 0;
    response_handler_command_valid = 0;
    unique if (load_resbuf) begin
      response_handler_command = RESBUF_GNT;
      response_handler_command_thread = arbited;
      response_handler_command_valid = 1;
    end
    else if (result_accept) begin
      response_handler_command = NONE;
      response_handler_command_thread = arbited_r;
      response_handler_command_valid = 1;
    end
    else begin
      response_handler_command_valid = 0;
    end
  end

  // Connect output to output register
  assign out_result = out_result_r;

  // Debug
  int i;

  always_comb begin
    active_threads = 0;
    for (i = 0; i < THREADS; i++) begin
      active_threads += (status_r[i] != IDLE);
    end
  end

endmodule