/**
 * The core.
 *
 * Implements the datapath and control for the core.
 */

module ray_datapath
  (
    // Outputs
    out_req_shift, out_result, out_result_valid, out_thread, out_status, out_status_valid, out_obj_addr, out_obj_read, out_instruction_address,
    // Inputs
    clk, reset_n, inp_req, inp_command, inp_obj_data, inp_instruction
  );

  import types_pkg::*;

  // Control
  input  logic                                reset_n;
  input  logic                                clk;

  // Request buffer interface
  input  logic [WORD_BITS               -1:0] inp_req;
  output logic                                out_req_shift;

  // Result buffer interface
  output logic [BYTE_BITS               -1:0] out_result;
  output logic                                out_result_valid;

  // Command and status interface
  output logic [$clog2(THREADS)         -1:0] out_thread;
  input  command_t                            inp_command;
  output status_t                             out_status;
  output logic                                out_status_valid;

  // Object buffer interface
  output logic [OBJ_BUFFER_ADDR_BITS    -1:0] out_obj_addr;
  output logic                                out_obj_read;
  input  logic [WORD_BITS               -1:0] inp_obj_data;
  
  // Instruction memory interface
  output logic [PC_BITS                 -1:0] out_instruction_address;
  input  instruction_union_t                  inp_instruction;

  //*****************************************************************************
  //
  // Thread generator
  //
  // Generates thread select signal
  // Generates valid signal
  //*****************************************************************************

  logic [$clog2(THREADS)                -1:0] s00_thread_r;
  logic                                       s00_thread_valid_r;
  
  logic [PC_BITS -1                       :0] pcs_r [0:THREADS -1];

  always_ff @(posedge clk or negedge reset_n) begin
    if ( !reset_n ) begin
      s00_thread_r <= 2'b11;
      s00_thread_valid_r <= 0;
    end
    else begin
      s00_thread_r <= s00_thread_r + 1;
      s00_thread_valid_r <= 1;
    end
  end

  //*****************************************************************************
  //
  // Pipeline step 0
  //
  // Getting current PC from array of PCs
  // Instantiating instruction memory
  //*****************************************************************************
  logic [$clog2(THREADS)                -1:0] s0_thread_r;
  logic                                       s0_thread_valid_r;

  logic [PC_BITS -1                       :0] s0_pc;
  logic [PC_BITS -1                       :0] s0_pc_r;

  instruction_union_t                         s0_instruction_r;

  assign s0_pc = pcs_r[s00_thread_r];

  always_ff @(posedge clk or negedge reset_n) begin
    if ( !reset_n ) begin
      s0_thread_valid_r <= 0;
    end
    else begin
      s0_thread_valid_r <= s00_thread_valid_r;
    end
  end

  always_ff @(posedge clk) begin
    s0_thread_r <= s00_thread_r;
    s0_pc_r <= s0_pc;
  end

  // instruction_memory IMEM (
  //   .clk(clk),
  //   .inp_addr(s0_pc),
  //   .out_instruction(s0_instruction_r)
  // );
  
  assign out_instruction_address = s0_pc;
  assign s0_instruction_r = inp_instruction;


  //*****************************************************************************
  //
  // Pipeline step 1
  //
  // Incrementing PC
  // Instantiating register file
  //*****************************************************************************
  logic [$clog2(THREADS)                -1:0] s1_thread_r;
  logic                                       s1_thread_valid_r;
  logic [PC_BITS -1                       :0] s1_pc_r;

  instruction_union_t                         s1_instruction_r;

  logic                                       regfile_we;
  logic [WORD_BITS                      -1:0] regfile_di_w;
  logic [REGISTER_ADDR_BITS             -1:0] regfile_addr_w;

  logic [WORD_BITS                      -1:0] s1_a_r;
  logic [WORD_BITS                      -1:0] s1_b_r;


  always_ff @(posedge clk or negedge reset_n) begin
    if ( !reset_n ) begin
      s1_thread_valid_r <= 0;
    end
    else begin
      s1_thread_valid_r <= s0_thread_valid_r;
    end
  end

  always_ff @(posedge clk) begin
    s1_thread_r <= s0_thread_r;
    s1_pc_r <= s0_pc_r + 1;
    s1_instruction_r <= s0_instruction_r;
  end

  register_file REGFILE (
    .clk(clk),
    .we(regfile_we),
    .di_w(regfile_di_w),
    .addr_a({s0_thread_r, s0_instruction_r.r.ra}),
    .addr_b({s0_thread_r, s0_instruction_r.r.rb}),
    .addr_w(regfile_addr_w),
    .do_a(s1_a_r),
    .do_b(s1_b_r)
  );


  //*****************************************************************************
  //
  // Pipeline step 2
  //
  // Muxing ALU/FPU operands
  // Instantiating ALU/FPU
  // Choosing next PC for thread dependent on branch conditions
  // Setting output signals thread, status and object buffer
  //*****************************************************************************
  logic [$clog2(THREADS)                -1:0] s2_thread_r;
  logic                                       s2_thread_valid_r;

  instruction_union_t                         s2_instruction_r;

  always_ff @(posedge clk or negedge reset_n) begin
    if ( !reset_n ) begin
      s2_thread_valid_r <= 0;
    end
    else begin
      s2_thread_valid_r <= s1_thread_valid_r;
    end
  end

  always_ff @(posedge clk) begin
    s2_thread_r <= s1_thread_r;
    s2_instruction_r <= s1_instruction_r;
  end

  logic                                       s2_alu_branch;
  logic [WORD_BITS                      -1:0] s2_alu_result_nxt;
  logic [WORD_BITS                      -1:0] s2_alu_result_r;

  alu ALU (
    .inp_a(s1_a_r),
    .inp_b(s1_b_r),
    .inp_instruction(s1_instruction_r),
    .inp_command(inp_command),
    .out_result(s2_alu_result_nxt),
    .out_branch(s2_alu_branch)
  );

  always_ff @(posedge clk) begin
    s2_alu_result_r <= s2_alu_result_nxt;
  end

  always_ff @(posedge clk or negedge reset_n) begin
    if ( !reset_n ) begin
      //for (int i=0; i<THREADS; i++) pcs_r[i] <= 0;
      pcs_r <= '{default:0};
    end
    else begin
      if ( s1_thread_valid_r ) begin
        pcs_r[s1_thread_r] <= ( s2_alu_branch ) ? s1_instruction_r.b.x : s1_pc_r;
      end
    end
  end

  assign out_thread = s1_thread_r;
  assign out_status = status_t'(s1_instruction_r.i.x[1:0]);
  assign out_status_valid = s1_thread_valid_r & (s1_instruction_r.r.op == STAT);
  assign out_obj_addr = s2_alu_result_nxt[OBJ_BUFFER_ADDR_BITS +1:2];
  assign out_obj_read = s1_thread_valid_r & (
                          (s1_instruction_r.r.op == LW) |
                          (s1_instruction_r.r.op == LB) |
                          (s1_instruction_r.r.op == LFX));


  logic [WORD_BITS                      -1:0] s3_fpu_result;

  fpu FPU (
    .clk(clk),
    .inp_a(s1_a_r),
    .inp_b(s1_b_r),
    .inp_instruction(s1_instruction_r),
    .out_result(s3_fpu_result)
  );


  //*****************************************************************************
  //
  // Pipeline step 3
  //
  // Muxing out correct byte from object buffer word
  // Converting FX1.7 to FP32
  // Setting up write to register file
  // Converting FP32 to UINT
  // Setting output signals
  //*****************************************************************************

  logic [WORD_BITS                      -1:0] s3_obj_buffer_byte;
  always_comb begin
    case (s2_alu_result_r[1:0])
      2'b00   : s3_obj_buffer_byte = {24'd0, inp_obj_data[1*BYTE_BITS -1:0*BYTE_BITS]};
      2'b01   : s3_obj_buffer_byte = {24'd0, inp_obj_data[2*BYTE_BITS -1:1*BYTE_BITS]};
      2'b10   : s3_obj_buffer_byte = {24'd0, inp_obj_data[3*BYTE_BITS -1:2*BYTE_BITS]};
      2'b11   : s3_obj_buffer_byte = {24'd0, inp_obj_data[4*BYTE_BITS -1:3*BYTE_BITS]};
      default : s3_obj_buffer_byte = {24'd0, inp_obj_data[4*BYTE_BITS -1:3*BYTE_BITS]};
    endcase
  end

  logic [WORD_BITS                      -1:0] s3_conv_from_fx;

  fx2fp FX2FP (
    .fxnum(s3_obj_buffer_byte[BYTE_BITS -1:0]),
    .fpnum(s3_conv_from_fx)
  );


  assign regfile_addr_w = {s2_thread_r, s2_instruction_r.r.rd};

  always_comb begin
    case (s2_instruction_r.r.op)
      ADD     : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s2_alu_result_r};
      SUB     : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s2_alu_result_r};
      SRL     : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s2_alu_result_r};
      LUI     : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s2_alu_result_r};
      ADDI    : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s2_alu_result_r};
      SUBI    : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s2_alu_result_r};
      MULI    : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s2_alu_result_r};
      LW      : {regfile_we, regfile_di_w} = {s2_thread_valid_r, inp_obj_data};
      LB      : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s3_obj_buffer_byte};
      LFX     : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s3_conv_from_fx};
      FADD    : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s3_fpu_result};
      FSUB    : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s3_fpu_result};
      FMUL    : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s3_fpu_result};
      BEQ     : {regfile_we, regfile_di_w} = {1'b0             , { WORD_BITS { 1'bx } }};
      BPOS    : {regfile_we, regfile_di_w} = {1'b0             , { WORD_BITS { 1'bx } }};
      BNEG    : {regfile_we, regfile_di_w} = {1'b0             , { WORD_BITS { 1'bx } }};
      BCEQ    : {regfile_we, regfile_di_w} = {1'b0             , { WORD_BITS { 1'bx } }};
      STAT    : {regfile_we, regfile_di_w} = {1'b0             , { WORD_BITS { 1'bx } }};
      POP     : {regfile_we, regfile_di_w} = {s2_thread_valid_r, inp_req};
      PUSH    : {regfile_we, regfile_di_w} = {1'b0             , { WORD_BITS { 1'bx } }};
      FADDI   : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s3_fpu_result};
      FMULI   : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s3_fpu_result};
      ADDUI   : {regfile_we, regfile_di_w} = {s2_thread_valid_r, s2_alu_result_r};
      default : {regfile_we, regfile_di_w} = {1'b0             , { WORD_BITS { 1'bx } }};
    endcase 
  end

  logic                                       s3_fp2fx_ovf;

  always_comb begin
    out_result_valid = (s2_instruction_r.r.op == PUSH) & s2_thread_valid_r;

    out_req_shift = s2_thread_valid_r & (s2_instruction_r.r.op == POP);

    s3_fp2fx_ovf = s3_fpu_result[31:23] != 9'b001111111;
    out_result = s3_fp2fx_ovf ? 8'hFF : s3_fpu_result[22:15];
  end


endmodule