# ------------------------------------------------------------------------------
# Author       : Oystein Gjermundnes
# Organization : Norwegian University of Science and Technology (NTNU)
#                http://www.iet.ntnu.no/en
# Course       : Design of digital systems 1 (DDS1)
# Year         : 2015
# Fork         : git clone git@bitbucket.org:oysteingjermundnes/dds1.git
# Description  : Raytracer model written in python
# License      : This is free and unencumbered software released into the 
#                public domain (UNLICENSE)
# ------------------------------------------------------------------------------
import sys
from PIL import Image
from numpy import *
from math import *
from rayman.color import *
from rayman.vector import *
from rayman.uvnframe import *
from rayman.material import *
from rayman.isectpoint import *
from rayman.ray import *
from rayman.sphere import *
from rayman.scene import *
import struct
from collections import Counter

import os
import pickle

from rayman.memory import *

from fpu.fp32_mul import fp32_mul
from fpu.fp32_add import fp32_add

# def chk_no_den_inf_nan(num):
#   f = as_int(num)
#   sfc = f & 0x7fffff
#   exp = (f >> 23) & 0xFF
#   if (exp == 0) and (sfc != 0):
#     return 'denormal'
#   elif exp == 0xFF:
#     return 'Inf' if (sfc == 0) else 'NaN'
#   else:
#     return ''


class Core:
  ##############################################################################
  ## Constructor for the scene
  #  
  #  The constructor does the following: 
  #  - Sets max recursion depth
  #  - Specifies the camera position
  #  - Spcifies abient color
  #  - Creates an empty frame buffer 
  # 
  # @param self The object pointer.
  # @param scene_file File name of the file representing the scene.
  ##############################################################################
  def __init__(self, objbuffer):    
    self.objbuffer = objbuffer
    self.max_reflection_recursion_depth = 2

    self.stats = []

  def save_stats(self, savepath):
    with open(savepath, 'wb') as fh:
      pickle.dump(self.stats, fh, protocol=2)

  def ray_request(self, x, y, direction):
    # cam_pos = Vector(0, 0, 0)     
    # cam_pos.x = self.objbuffer.buffer.load_word(3*4)
    # cam_pos.y = self.objbuffer.buffer.load_word(4*4)
    # cam_pos.z = self.objbuffer.buffer.load_word(5*4)
    # color = self.raytrace_tail(cam_pos, direction)

    self.x = x
    self.y = y
    color = self.raytrace_tail_ass(direction)

    return (x, y, color)


##############################################################################
  ## The raytrace function computes the color of a ray.
  #
  # If the ray hits an object and this object is partly transparent or has
  # a surface that reflects rays, then a reflection and a refraction ray is created.
  # The color of the ray is the blend of the color at the point of intersection
  # and the color of the refracted and reflected rays.
  #
  # The color of the reflected and refracted rays are computed through recursive
  # calls to this function.
  #
  # @param self The object pointer.
  # @param Ri Incoming ray.
  # @param recursion_depth The recursion depth of the incoming ray.
  ##############################################################################
  def raytrace_tail(self, origin, direction):
    color = Color(0.0, 0.0, 0.0)
    rec_surface = 0xFFFF0000
    k_refl = 1.0

    for _ in range(self.max_reflection_recursion_depth+1):  # 'raytraceloop' in assembly
      # Check if the ray collides with any primitive
      isect_surface = 0xFF # Eller annen "stor" konstant

      # Set the distance to the intersection point to a large value initially.
      min_t = 10000000  # Could be implemented as a do-while structure in assembly?

      num_spheres = self.objbuffer.buffer.load_word(6*4)

      for p in range(num_spheres):
        if p != rec_surface:
          
          center = Vector(0,0,0)
          center.x = self.objbuffer.buffer.load_word((7+6*p+0)*4)
          center.y = self.objbuffer.buffer.load_word((7+6*p+1)*4)
          center.z = self.objbuffer.buffer.load_word((7+6*p+2)*4)
          radius_squared = self.objbuffer.buffer.load_word((7+6*p+3)*4)

        
          u   = direction
          dP  = center.sub(origin)
          udP = u.dot(dP)
          A   = dP.sub(u.mult(udP))
          D   = radius_squared - A.dot(A)
          # No intersection
          if (D < 0):
            continue
            
          sqrtD = sqrtapprox_a(D)

          t = udP - sqrtD

          if(t < 0):
            t = udP + sqrtD

          if(t < 0):
            # No intersections found.
            continue
          
          # Found closer intersection point
          if(t < min_t):
            min_t = t
            isect_surface = p
            
      if (isect_surface == 0xFF):
        break

      origin = origin.add(direction.mult(min_t))

      p = isect_surface
      radius_inverse = self.objbuffer.buffer.load_word((7+6*p+4)*4)
      material_id = self.objbuffer.buffer.load_word((7+6*p+5)*4)
      center = Vector(0,0,0)
      center.x = self.objbuffer.buffer.load_word((7+6*p+0)*4)
      center.y = self.objbuffer.buffer.load_word((7+6*p+1)*4)
      center.z = self.objbuffer.buffer.load_word((7+6*p+2)*4)
      
      isect_normal  = origin.sub(center).mult(radius_inverse)

    # Compute the color
      local_color = self.calculate_local_color(origin, direction, isect_normal, isect_surface)
      color = color.add(local_color.scalar_mult(k_refl))

      #color = self.calculate_local_color_ass(origin, direction, isect_normal, isect_surface, k_refl, color)
      
      # Only used for getting next k_refl
      material_arr_ptr = (7+6*num_spheres)*4
      k_refl *= self.objbuffer.buffer.load_byte(material_arr_ptr+(4*material_id+3)*4) * 2**-7

      # Trace reflected rays
      if(k_refl < 0.05 ):
        break
      rec_surface = isect_surface  # Kan flyttes opp for å spare et register under fargeutregning!

      # Create reflection
      Ri_dot_N = direction.dot(isect_normal)
      direction = direction.sub(isect_normal.mult(np.float32(2)*Ri_dot_N))

    return color


  def raytrace_tail_ass(self, direction):
    regs = {'0': 0}
    with open('rayman/program.txt') as fh:
      lines = []
      labels = {}
      i = 0
      for line in fh.readlines():
        line = line.split(';')[0].strip()
        if line:
          if ':' in line:
            labels[line.split(':')[0]] = i
          else:
            lines.append(line)
            i += 1

      # for i, line in enumerate(lines):
      #   try:
      #     label = [key for key,value in labels.items() if value==i][0]
      #   except:
      #     label = ''
      #   print('{:03d}: {:15s}: {}'.format(i, label, line))
      # raise 'yo'

      cycles = 0
      history = []
      instructions = Counter()
      i = 0

      status = 0b00 # idle
      command = 0b01 # start
      resbuf = []
      pops = 0
      while i < (len(lines) - 1): # NOT executing last instruction (loop)
        line = lines[i]
        history.append(i)
        cycles += 1

        i += 1

        s = line.replace(',','').split()

        instructions.update([s[0]])

        # floatflag = ''


        if s[0] == 'nop':
          continue

        dest = s[1][1:]

        if s[0] == 'lw':
          loc = s[2].split('[$')
          imm = int(loc[0])
          reg = loc[1][:-1]
          addr = regs[reg] + imm
          regs[dest] = self.objbuffer.buffer.load_word(addr)

        elif s[0] == 'pop':
          if pops == 0:
            regs[dest] = direction.z
          elif pops == 1:
            regs[dest] = direction.y
          elif pops == 2:
            regs[dest] = direction.x
          else:
            print("EASY ON THE POPPIN'!")
          pops += 1

        elif s[0] == 'push':
          a = regs[s[1][1:]]
          if(type(a) in (np.int64, np.int32, int)):
            a = as_float(a)
          b = np.float32(1.0)
          res_int = as_int(fp32_add(a, b, 0))
          intified = (res_int >> 15) & 0xFF
          overflowing = (res_int >> 23) != 0b001111111
          a_uint = intified if not overflowing else 0xFF
          resbuf.append(a_uint)

        elif s[0] == 'stat':
          status = int(s[1])
          if status == 2: # 'resbuf_req'
            command = 2 # 'resbuf_gnt'

        elif s[0] == 'bceq':
          y = int(s[1])
          if y == command:
            i = labels[s[2]]

        elif s[0] == 'debug':
          print('${}: {} ({})'.format(dest, regs[dest], type(regs[dest])))

        elif s[0] == 'beq':
          a = regs[s[1][1:]]
          b = regs[s[2][1:]]
          assert(type(a) == type(b))
          if a == b:
            i = labels[s[3]]

        elif s[0] == 'goto':
          i = labels[s[1]]

        elif s[0] == 'bneg':
          a = regs[s[1][1:]]
          if a < 0:
            i = labels[s[2]]

        elif s[0] == 'bpos':
          a = regs[s[1][1:]]
          if a >= 0:
            i = labels[s[2]]

        elif s[0] == 'lb':
          loc = s[2].split('[$')
          imm = int(loc[0])
          reg = loc[1][:-1]
          addr = regs[reg] + imm
          regs[dest] = self.objbuffer.buffer.load_byte(addr)

        elif s[0] == 'lfx':
          loc = s[2].split('[$')
          imm = int(loc[0])
          reg = loc[1][:-1]
          addr = regs[reg] + imm
          fx = self.objbuffer.buffer.load_byte(addr)
          regs[dest] = float32(min(1.0, float32(fx)/0x80))
          # floatflag = chk_no_den_inf_nan(regs[dest])

        elif s[0] == 'muli':
          a = regs[s[2][1:]]
          imm = int(s[3])
          assert(type(a) in (np.int64, np.int32, int))
          regs[dest] = a * imm

        elif s[0] == 'fsub':
          a = regs[s[2][1:]]
          b = regs[s[3][1:]]
          if(type(a) in (np.int64, np.int32, int)):
            a = as_float(a)
          if(type(b) in (np.int64, np.int32, int)):
            b = as_float(b)
          # regs[dest] = a - b
          regs[dest] = fp32_add(a, b, 1)
          # floatflag = chk_no_den_inf_nan(regs[dest])

        elif s[0] == 'fadd':
          a = regs[s[2][1:]]
          b = regs[s[3][1:]]
          if(type(a) in (np.int64, np.int32, int)):
            a = as_float(a)
          if(type(b) in (np.int64, np.int32, int)):
            b = as_float(b)
          # regs[dest] = a + b
          regs[dest] = fp32_add(a, b, 0)
          # floatflag = chk_no_den_inf_nan(regs[dest])

        elif s[0] == 'fmul':
          a = regs[s[2][1:]]
          b = regs[s[3][1:]]
          if(type(a) in (np.int64, np.int32, int)):
            a = as_float(a)
          if(type(b) in (np.int64, np.int32, int)):
            b = as_float(b)
          # regs[dest] = a * b
          regs[dest] = fp32_mul(a, b)
          # floatflag = chk_no_den_inf_nan(regs[dest])

        elif s[0] == 'faddi':
          a = regs[s[2][1:]]
          b = int(s[3], 0) << 16
          if(type(a) in (np.int64, np.int32, int)):
            a = as_float(a)
          if(type(b) in (np.int64, np.int32, int)):
            b = as_float(b)
          # regs[dest] = a + b
          regs[dest] = fp32_add(a, b, 0)
          # floatflag = chk_no_den_inf_nan(regs[dest])

        elif s[0] == 'fmuli':
          a = regs[s[2][1:]]
          b = int(s[3], 0) << 16
          if(type(a) in (np.int64, np.int32, int)):
            a = as_float(a)
          if(type(b) in (np.int64, np.int32, int)):
            b = as_float(b)
          # regs[dest] = a * b
          regs[dest] = fp32_mul(a, b)
          # floatflag = chk_no_den_inf_nan(regs[dest])

        elif s[0] == 'srl':
          a = regs[s[2][1:]]
          if(type(a) in (np.float64, np.float32, float)):
            a = as_int(a)
          regs[dest] = a >> 1

        elif s[0] == 'lui':
          imm = eval(s[2])
          assert(type(imm) in (np.int64, np.int32, int))
          assert(imm <= 0xFFFF)
          regs[dest] = imm << 16

        elif s[0] == 'ori':
          a = regs[s[2][1:]]
          imm = eval(s[3])
          assert(type(a) in (np.int64, np.int32, int))
          assert(imm <= 0xFFFF)
          regs[dest] = a | imm

        elif s[0] == 'addi':
          a = regs[s[2][1:]]
          imm = eval(s[3])
          assert(type(a) in (np.int64, np.int32, int))
          assert(imm <= 0xFFFF)
          regs[dest] = a + imm

        elif s[0] == 'addui':
          a = regs[s[2][1:]]
          imm = eval(s[3])
          assert(type(a) in (np.int64, np.int32, int))
          assert(imm <= 0xFFFF)
          regs[dest] = a + (imm<<16)

        elif s[0] == 'subi':
          a = regs[s[2][1:]]
          imm = eval(s[3])
          assert(type(a) in (np.int64, np.int32, int))
          assert(imm <= 0xFFFF)
          regs[dest] = a - imm

        elif s[0] == 'add':
          a = regs[s[2][1:]]
          b = regs[s[3][1:]]
          if(type(a) in (np.float64, np.float32, float)):
            a = as_int(a)
          if(type(b) in (np.float64, np.float32, float)):
            b = as_int(b)
          regs[dest] = a + b

        elif s[0] == 'sub':
          a = regs[s[2][1:]]
          b = regs[s[3][1:]]
          if(type(a) in (np.float64, np.float32, float)):
            a = as_int(a)
          if(type(b) in (np.float64, np.float32, float)):
            b = as_int(b)
          regs[dest] = a - b

        else: 
          print(s)

        # if floatflag:
        #   print('-'*5)
        #   print('Got {}'.format(floatflag))
        #   print(' '.join(s))
        #   print('a: {} {} b: {} = {}'.format(a, s[0], b, regs[dest]))
        #   print('-'*5)

        # if floatflag == 'denormal':
        #   regs[dest] = np.float32(0.0)  # Flush to zero
        #   print('Flushing to zero')
        # elif floatflag:
        #   print('-'*5)
        #   print('Got {}'.format(floatflag))
        #   print(' '.join(s))
        #   print('a: {} {} b: {} = {}'.format(a, s[0], b, regs[dest]))
        #   print('-'*5)


    if status != 0b00:
      print('Ohno! We finished with status: {:02b}'.format(status))

    self.stats.append({'position': (self.x, self.y),
                       'cycles': cycles,
                       'history': history,
                       'instructions': dict(instructions)})



    return (resbuf[0], resbuf[1], resbuf[2])


def as_float(a):
    return struct.unpack('<f', struct.pack('<l', np.int32(a)))[0]


def as_int(a):
    return np.int32(struct.unpack('<l', struct.pack('<f', a))[0])