# -----------------------------------------------------------------------------
# Author       : Oystein Gjermundnes
# Organization : Norwegian University of Science and Technology (NTNU)
#                http://www.iet.ntnu.no/en
# Course       : Design of digital systems 1 (DDS1)
# Year         : 2015
# Fork         : git clone git@bitbucket.org:oysteingjermundnes/dds1.git
# Description  : Raytracer model written in python
# License      : This is free and unencumbered software released into the
#                public domain (UNLICENSE)
# -----------------------------------------------------------------------------

from rayman.scene import *
import math

class Cpu():
    ##########################################################################
    # Constructor
    #
    # @param self      Pointer to the scene.
    # @param filename  Name of the file containing the scene description.
    ##########################################################################

    def __init__(self, filename, ram, raymanager):

        self.scene = Scene(filename)
        self.scene.load()

        self.ram = ram

        self.start_addr = 0x80000000
        self.current_addr = self.start_addr

        self.write_scene_to_ram()

        raymanager.write_register(self.start_addr, 0x8)
        raymanager.write_register(1, 0x0)  # RUN

    def write_scene_to_ram(self):
        self.write_word_and_advance(self.scene.light_pos.x)   # W0
        self.write_word_and_advance(self.scene.light_pos.y)   # W1
        self.write_word_and_advance(self.scene.light_pos.z)   # W2

        self.write_word_and_advance(self.scene.cam_pos.x)     # W3
        self.write_word_and_advance(self.scene.cam_pos.y)     # W4
        self.write_word_and_advance(self.scene.cam_pos.z)     # W5

        self.write_word_and_advance(self.scene.cam_target.x)  # W6
        self.write_word_and_advance(self.scene.cam_target.y)  # W7
        self.write_word_and_advance(self.scene.cam_target.z)    # W8

        self.write_word_and_advance(self.scene.cam_fov)         # W9

        self.write_halfword_and_advance(self.scene.w_pix)       # W10
        self.write_halfword_and_advance(self.scene.h_pix)       # W10

        sphere_size = len(self.scene.spheres)
        material_size = len(self.scene.materials)

        sphere_pointer = self.current_addr + 4 * 4
        material_pointer = sphere_pointer + 6 * 4 * sphere_size

        self.write_word_and_advance(sphere_pointer)             # W11
        self.write_word_and_advance(sphere_size)                # W12

        self.write_word_and_advance(material_pointer)           # W13
        self.write_word_and_advance(material_size)              # W14

        material_names = list(self.scene.materials.keys())

        for sphere in self.scene.spheres:
            self.write_word_and_advance(sphere.center.x)        # W0
            self.write_word_and_advance(sphere.center.y)        # W1
            self.write_word_and_advance(sphere.center.z)        # W2

            self.write_word_and_advance(sphere.radius_squared)  # W3
            self.write_word_and_advance(sphere.radius_inverse)  # W4

            material_id = material_names.index(sphere.material_id)
            self.write_word_and_advance(material_id)          # W5

        for material_name in material_names:
            material = self.scene.materials[material_name]

            self.write_fx1_7_and_advance(material.ac.r)       # W0
            self.write_fx1_7_and_advance(material.ac.g)       # W0
            self.write_fx1_7_and_advance(material.ac.b)       # W0
            self.write_byte_and_advance(0)                    # W0

            self.write_fx1_7_and_advance(material.dc.r)       # W1
            self.write_fx1_7_and_advance(material.dc.g)       # W1
            self.write_fx1_7_and_advance(material.dc.b)       # W1
            self.write_byte_and_advance(0)                    # W1

            self.write_fx1_7_and_advance(material.sc.r)       # W2
            self.write_fx1_7_and_advance(material.sc.g)       # W2
            self.write_fx1_7_and_advance(material.sc.b)       # W2
            self.write_byte_and_advance(0)                    # W2

            self.write_fx1_7_and_advance(material.k_refl)     # W3
            shininesspowapp = int(round(math.log2(material.shininess)))
            self.write_byte_and_advance(shininesspowapp)   # W3
            self.write_halfword_and_advance(0)               # W3

    def write_word_and_advance(self, data):
        self.ram.store_word(data, self.current_addr)
        self.current_addr += 4

    def write_halfword_and_advance(self, data):
        self.ram.store_halfword(data, self.current_addr)
        self.current_addr += 2

    def write_byte_and_advance(self, data):
        self.ram.store_byte(data, self.current_addr)
        self.current_addr += 1

    def write_fx1_7_and_advance(self, data):
        # Clamp and scale
        fx1_7 = int(min(max(data, 0), 1) * 128)
        self.write_byte_and_advance(fx1_7)
