# -----------------------------------------------------------------------------
# Author       : Oystein Gjermundnes
# Organization : Norwegian University of Science and Technology (NTNU)
#                http://www.iet.ntnu.no/en
# Course       : Design of digital systems 1 (DDS1)
# Year         : 2015
# Fork         : git clone git@bitbucket.org:oysteingjermundnes/dds1.git
# Description  : Raytracer model written in python
# License      : This is free and unencumbered software released into the
#                public domain (UNLICENSE)
# -----------------------------------------------------------------------------

import numpy as np
import struct
from numpy import float32, float64


class Memory():

    def __init__(self):
        self.memory = {}

    def word_align_address(self, address):
        return address - (address % 4)

    def half_word_align_address(self, address):
        return address - (address % 2)

    def assert_word_aligned(self, address):
        assert self.word_align_address(address) == address

    def assert_half_word_aligned(self, address):
        assert self.half_word_align_address(address) == address

    def store_word(self, data, address):
        self.assert_word_aligned(address)
        self.memory[address] = data

    def load_word(self, address, disable_warn=False):
        self.assert_word_aligned(address)
        try:
            return self.memory[address]
        except KeyError as e:
            if not disable_warn:
                print('Load from empty RAM location: 0x{:08x}'.format(address))
            return 0

    def load_byte(self, address):
        wa_address = self.word_align_address(address)
        byte_i = address % 4

        word = self.load_word(wa_address)
        return 0xFF & (word >> (byte_i * 8))

    def store_byte(self, data, address):
        assert data == data & 0xFF
        wa_address = self.word_align_address(address)
        byte_i = address % 4

        word = self.load_word(wa_address, True)

        mask = 0xFF << (byte_i * 8)
        word = (~mask & word) | (data << (byte_i * 8))
        self.store_word(word, wa_address)

    def store_halfword(self, data, address):
        assert data == data & 0xFFFF
        self.assert_half_word_aligned(address)
        self.store_byte(data & 0x00FF, address)
        self.store_byte((data & 0xFF00) >> 8, address + 1)

    def load_halfword(self, address):
        self.assert_half_word_aligned(address)
        data = self.load_byte(address)
        data |= self.load_byte(address + 1) << 8
        return data

    def __str__(self):
        string = ''
        for address in sorted(self.memory):
            value = self.memory[address]
            string += '0x{:08x}: {}\n'.format(address, value)
        string = string[0:-1]
        return string

    def dump(self, filename):
        result = ''
        maxlen = len(hex(max(self.memory))) - 2
        for address in range(0, max(self.memory) + 1, 4):
            word = self.load_word(address, False)
            wordword = word
            if type(word) in (float, float32, float64):
                word = as_int(word)
            result += '{:08x} // 0x{:0{}x} - {} \n'.format(word, address, maxlen, wordword)

        with open(filename, 'w') as fh:
            fh.write(result)


def as_int(a):
    return np.uint32(struct.unpack('<l', struct.pack('<f', a))[0])


if __name__ == '__main__':
    ram = Memory()
    ram.store_word(0xDEADBEEF, 0)
    ram.store_byte(0x00, 1)
    print('{:08x}'.format(ram.load_word(0)))
    print('{:02x}'.format(ram.load_byte(1)))

    ram.store_halfword(0xfaf1, 6)
    print('{:08x}'.format(ram.load_halfword(6)))

    ram.store_word(1337, 8)

    print(ram)
    ram.dump('yofile.dat')
