# ------------------------------------------------------------------------------
# Author       : Oystein Gjermundnes
# Organization : Norwegian University of Science and Technology (NTNU)
#                http://www.iet.ntnu.no/en
# Course       : Design of digital systems 1 (DDS1)
# Year         : 2015
# Fork         : git clone git@bitbucket.org:oysteingjermundnes/dds1.git
# Description  : Raytracer model written in python
# License      : This is free and unencumbered software released into the
#                public domain (UNLICENSE)
# ------------------------------------------------------------------------------
import sys
from PIL import Image
from numpy import *
from math import *
from rayman.color import *
from rayman.vector import *
from rayman.uvnframe import *
from rayman.material import *
from rayman.isectpoint import *
from rayman.ray import *
from rayman.sphere import *
from rayman.scene import *


import datetime
import sys
import os
import subprocess

from rayman.memory import *


import struct
def as_int(a):
    return np.uint32(struct.unpack('<l', struct.pack('<f', a))[0])


class Raymanager:

    def __init__(self, ram, objbuffer, core):
        self.registers = Memory()

        self.ram = ram

        self.scene_buffer = Memory()

        self.objbuffer = objbuffer

        self.core = core


    def write_register(self, data, address):
        self.registers.store_word(data, address)
        if (address == 0):
            if (data == 0):  # COMMAND_NOP
                pass
            elif (data == 1):  # COMMAND_RUN
                self.run()
            elif (data == 2):  # COMMAND_STOP
                pass
            elif (data == 3):  # COMMAND_CLEAR_IRQ
                pass

    def run(self):
        scene_address = self.registers.load_word(0x8)
        for i in range(15):
            self.scene_buffer.store_word(self.ram.load_word(scene_address+i*4), i*4)

        print(self.scene_buffer)
        
        response = self.objbuffer.read_req(scene_address) # Make obj buffer load its stuff
        assert response == 'DONE'

        self.create_uvn_frame()

        self.w_pix = self.scene_buffer.load_halfword(10*4)
        self.h_pix = self.scene_buffer.load_halfword(10*4 + 2)

        self.a = self.w_pix/self.h_pix

        ## Create a framebuffer
        self.img    = Image.new('RGB', (self.w_pix, self.h_pix),'green')
        ## 2D array of pixels
        self.pixels = self.img.load()

        # Make cores work!
        self.render()

        self.write_fb()
    
    

    def create_uvn_frame(self):
        self.cam_pos = Vector(0, 0, 0)     
        self.cam_pos.x = cam_fov = self.scene_buffer.load_word(3*4)
        self.cam_pos.y = cam_fov = self.scene_buffer.load_word(4*4)
        self.cam_pos.z = cam_fov = self.scene_buffer.load_word(5*4)

        self.cam_target = Vector(0, 0, 0)
        self.cam_target.x = cam_fov = self.scene_buffer.load_word(6*4)
        self.cam_target.y = cam_fov = self.scene_buffer.load_word(7*4)
        self.cam_target.z = cam_fov = self.scene_buffer.load_word(8*4)

        self.uvn_frame = uvnFrame(self.cam_pos, self.cam_target)

        print(self.uvn_frame)


    ##########################################################################
    # Display the rendered image
    #
    # Displays the rendered image using the default image viewer on the platform
    #
    # @param self The object pointer.
    ##########################################################################
    def write_fb(self):
        savedir = 'results'
        if not os.path.exists(savedir):
            os.makedirs(savedir)
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H.%M.%S") 
        filename = timestamp + '.png'
        savepath = os.path.join(savedir, filename)
        self.img.save(savepath)
        #self.img.show()

        imageViewerFromCommandLine = {'linux' :'eog',
                                      'win32' :'explorer',
                                      'darwin':'open'}[sys.platform]
        subprocess.run([imageViewerFromCommandLine, savepath])

        filename = timestamp + '.pickle'
        savepath = os.path.join(savedir, filename)
        self.core.save_stats(savepath)

    ##########################################################################
    # Render the scene
    #
    # Assumptions:
    #  - The rays are shot towards each pixel.
    #  - UVN camera coordinate system
    #  - Symmetrical view frustum centered at the optical axis
    #  - Ideal pinhole camera model:
    #    * Focal distance d
    #    * Aspect ratio a=w/h
    #    * Pixels are square image areas
    #
    # Reference:
    # [2] p 546, 539
    #
    # @param self The object pointer.
    ##########################################################################
    def render(self):
        # Pick a distance to the screen
        d = 10

        # Half the horizontal viewing angle
        cam_fov = self.scene_buffer.load_word(9*4)
        phi = cam_fov / 360 * 2 * math.pi

        # Half width and height
        half_w = d * math.tan(phi)
        half_h = half_w / self.a

        # The horizontal and vertical size of one pixel in world units
        pix_h_size = (2 * half_w / self.w_pix)
        pix_w_size = (2 * half_h / self.h_pix)

        # The center of pixel (i,j) in pixel coordinates is
        # (i+0.5, j+0.5)
        #
        # Displacement of the pixel (i,j) from the image center
        # in world units:
        #
        # Horizontal displacement: (i+0.5)*pix_w_size - half_w
        # Vertical displacement  : half_h - (j+0.5)*pix_h_size
        #
        # The displacement will be in the direction of the u and v
        # vectors of the camera

        # Position of the point in the middle of the screen
        P_s = self.cam_pos.add(self.uvn_frame.n.mult(-d))

        # u and v vector:
        u = self.uvn_frame.u
        v = self.uvn_frame.v

        requests_fh  = open('../verilog/requests.dat', 'w')
        responses_fh = open('../verilog/responses.dat', 'w')

        reqfmt  = '0_{:010b}_{:09b}_{:032b}_{:032b}_{:032b}\n'
        respfmt = '0_{:010b}_{:09b}_{:08b}_{:08b}_{:08b}\n'

        # Loop through all points in in the screen
        for j in range(self.h_pix):
            for i in range(self.w_pix):

                # Set initial recursion depth
                recursion_depth = 0

                # Position of the pixel on the screen
                R_p = P_s.add(u.mult((i + 0.5) * pix_w_size - half_w))
                R_p = R_p.add(v.mult(half_h - (j + 0.5) * pix_h_size))

                # Direction of the ray
                R_d = R_p.sub(self.cam_pos).normalize()
                R = Ray(self.cam_pos, R_d, "initial ray")

                # Compute the color for the ray using recursive
                # raytracing
                #color = self.raytrace(R, recursion_depth)
                x, y, color = self.core.ray_request(i, j, R_d)


                requests_fh.write(reqfmt.format(i,
                                                j,
                                                as_int(R_d.x),
                                                as_int(R_d.y),
                                                as_int(R_d.z)))

                responses_fh.write(respfmt.format(i,
                                                  j,
                                                  color[0],
                                                  color[1],
                                                  color[2]))

                # Store the color for the pixel
                self.pixels[x, y] = color
            # Print out percentage completeness information
            print("Progress: %d\r" % (j / self.h_pix * 100), end='\r')
            sys.stdout.flush()
