# ------------------------------------------------------------------------------
# Author       : Oystein Gjermundnes
# Organization : Norwegian University of Science and Technology (NTNU)
#                http://www.iet.ntnu.no/en
# Course       : Design of digital systems 1 (DDS1)
# Year         : 2015
# Fork         : git clone git@bitbucket.org:oysteingjermundnes/dds1.git
# Description  : Raytracer model written in python
# License      : This is free and unencumbered software released into the 
#                public domain (UNLICENSE)
# ------------------------------------------------------------------------------

from rayman.vector import *
from rayman.ray import *
from rayman.sphere import *
import numpy as np

################################################################################
## Class for parsing Scene description files and storing Scene descriptions.
#
# This class parses Scene descriptions. These Scene descriptions contains 
# information about the position and material.Material properties of 
# primitive.Primitive such as sphere.Sphere and plane.Plane. The format of the
# file representing the scene must be as follows:
#
# \verbatim
#
# # FEATURE ENABLES
# # reflection refraction spheres planes antialiasing_4X
# f 1          1          1       1      1
# 
# # LIGHT (only one)
# # position(X,Y,Z)
# l -1000.00 1000.00 1000.00
# 
# # CAMERA (only one)
# # position(X,Y,Z)      field_of_view(DEG)   look_at_point(X,Y,Z)
# c 00.00  00.00 10.00   25                   00.00 00.00 -10.00
# 
# # SCREEN RESOLUTION
# # width   height
# r 600     600
# 
# # DEBUG PIXEL
# # pixel_position(i,j)
# d 30  40
# 
# # SPHERES
# # position(X,Y,Z)       radius    material_id
# s -06.00 -03.00 -20.00  5.00      gold
# s  06.00 -03.00 -20.00  5.00      ruby
# s  00.00  03.00 -10.00  3.00      emerald
# 
# # PLANES (Ax + By + Cz + D = 0)
# # plane (A, B, C, D)              material_id
# p 00.00  01.00  00.00  10.00      chess
# p 00.00  00.00  01.00  30.00      wall
# 
# # MATERIALS
# # The materials will be enumerated in the order of appearance
# # material_id ambient_color(RGB)  diffuse_color(RGB)  specular_color(RGB)  alpha  shininess  k_refl  k_refr color_func
# m ruby        0.17 0.01 0.01      0.61 0.04 0.04      0.73 0.63 0.63       0.55   76         0.45    0.55   0
# m emerald     0.02 0.17 0.02      0.08 0.61 0.08      0.63 0.73 0.63       0.55   76         0.45    0.55   0
# m gold        0.25 0.22 0.06      0.35 0.31 0.09      0.80 0.72 0.21       1.00   83         0.35    0.00   0
# m chess       0.00 0.00 0.00      0.00 0.00 0.00      0.40 0.40 0.40       1.00   50         0.15    0.00   1
# m wall        0.00 0.00 0.00      0.00 0.00 0.00      0.00 0.00 0.00       1.00    0         0.00    0.00   0
#
# \endverbatim
#
#
################################################################################
class Scene():
  ##############################################################################
  ## Constructor
  #
  # @param self      Pointer to the scene.
  # @param filename  Name of the file containing the scene description.
  ##############################################################################
  def __init__(self, filename):
    ## Name of Scene file
    self.filename = filename

    ## Light position
    self.light_pos = Vector(0,0,0)
   
    ## Width of framebuffer
    self.w_pix = 0
    ## Height of framebuffer
    self.h_pix = 0

    ## Camera position [vector.Vector]
    self.cam_pos = Vector(0,0,0)
    ## Camera field of view [DEG]
    self.cam_fov = np.float32(0.0)
    ## Camera target [vector.Vector]
    self.cam_target = Vector(0,0,0)

    ## Array of primitives in the Scene
    self.spheres    = []

    ## Dictionary of Materials in the Scene.
    self.materials = {}

  ##############################################################################
  ## Load a scene
  # 
  # This function parses every line of the Scene file and checks the first 
  # character of each line in order to interpret what the line represents.
  #
  # @param self  Pointer to the Scene.
  ##############################################################################
  def load(self):
    file = open(self.filename)
    lines = file.readlines()
    file.close()

    for line in lines:
      if(len(line) > 0):
        token = line[0]
        line = line[1:]
        p = line.split()
        if(token == "l"):
          self.parse_light(p)
        elif(token == "c"):
          self.parse_camera(p)
        elif(token == "r"):
          self.parse_resolution(p)
        elif(token == "s"):
          self.parse_sphere(p)
        elif(token == "m"):
          self.parse_material(p)


  ##############################################################################
  ## Parse light
  # 
  # Function for parsing lights on the form:
  # \verbatim
  #
  # # LIGHT (only one)
  # # position(X,Y,Z)
  # l -1000.00 1000.00 1000.00
  #
  # \endverbatim
  # 
  # @param self Pointer to the Scene.
  # @param p    Tokens representing the light.
  ##############################################################################
  def parse_light(self, p):
    x = np.float32(p[0])
    y = np.float32(p[1])
    z = np.float32(p[2])
    self.light_pos = Vector(x,y,z)
    print("INFO: Light placed at (x, y, z) = (%f, %f, %f)" % (x, y, z))

  ##############################################################################
  ## Parse camera
  #
  # Function for parsing the camera position and orientation on the form:
  # \verbatim
  #
  # # CAMERA (only one)
  # # position(X,Y,Z)      field_of_view(DEG)   look_at_point(X,Y,Z)
  # c 00.00  00.00 10.00   25                   00.00 00.00 -10.00
  #
  # \endverbatim
  # @param self Pointer to the Scene.
  # @param p    Tokens representing the camera.
  ##############################################################################
  def parse_camera(self, p):
    cam_x = np.float32(p[0])
    cam_y = np.float32(p[1])
    cam_z = np.float32(p[2])
    self.cam_pos = Vector(cam_x, cam_y, cam_z)
    print("INFO: Camera placed at position (x,y,z) = (%f, %f, %f)" % (cam_x, cam_y, cam_z))

    self.cam_fov = np.float32(p[3])
    print("INFO: Camera field of view set to: %f DEG" % self.cam_fov)

    target_x = np.float32(p[4])
    target_y = np.float32(p[5])
    target_z = np.float32(p[6])
    self.cam_target = Vector(target_x, target_y, target_z)
    print("INFO: Camera pointed in the direction of target at (x,y,z) = (%f, %f, %f)" % (target_x, target_y, target_z))

  ##############################################################################
  ## Parse resolution
  # 
  # Function for parsing the light position on the form:
  # 
  # \verbatim
  # # SCREEN RESOLUTION
  # # width   height
  # r 600     600
  #
  # \endverbatim
  #
  # @param self Pointer to the Scene.
  # @param p    Tokens representing the camera.
  ##############################################################################
  def parse_resolution(self, p):
    self.w_pix = int(p[0])
    self.h_pix = int(p[1])
    print("INFO: Screen resolution set to WxH = %dx%d" % (self.w_pix, self.h_pix))


  ##############################################################################
  ## Parse sphere
  # 
  # Function for parsing sphere.Sphere primitive.Primitive on the form:
  # 
  # \verbatim
  #
  # # SPHERES
  # # position(X,Y,Z)       radius    material_id
  # s -03.00 -03.00 -20.00  5.00      gold
  # s -03.00 -03.00 -20.00  5.00      ruby
  # s -03.00 -03.00 -20.00  5.00      emerald
  # 
  # \endverbatim
  #
  # @param self Pointer to the Scene.
  # @param p    Tokens representing the sphere.Sphere.
  ##############################################################################
  def parse_sphere(self, p):
    pos_x       = np.float32(p[0])
    pos_y       = np.float32(p[1])
    pos_z       = np.float32(p[2])
    pos         = Vector(pos_x, pos_y, pos_z)
    radius      = np.float32(p[3])
    material_id = p[4]
    sphere      = Sphere(pos, radius, material_id) 
    self.spheres.append(sphere)
    print("INFO: Sphere of material %s with radius %f located at position (%f, %f, %f)" % (material_id, radius, pos_x, pos_y, pos_z))


  ##############################################################################
  ## Parse material
  # 
  # Function for parsing material.Material on the form: 
  # 
  # \verbatim
  #
  # # MATERIALS
  # # The materials will be enumerated in the order of appearance
  # # material_id ambient_color(RGB)  diffuse_color(RGB)  specular_color(RGB)  shininess  k_refl
  # m ruby        0.17 0.01 0.01      0.61 0.04 0.04      0.73 0.63 0.63       76         0.45  
  # m emerald     0.02 0.17 0.02      0.08 0.06 0.08      0.63 0.73 0.63       76         0.45  
  # m gold        0.25 0.22 0.06      0.35 0.31 0.09      0.80 0.72 0.21       83         0.45  
  #
  # \endverbatim
  #
  # @param self Pointer to the Scene.
  # @param p    Tokens representing the material.Material.
  ##############################################################################
  def parse_material(self, p):
    material_id  = p[0]
    ac_r         = np.float32(p[1]) 
    ac_g         = np.float32(p[2]) 
    ac_b         = np.float32(p[3]) 
    dc_r         = np.float32(p[4]) 
    dc_g         = np.float32(p[5]) 
    dc_b         = np.float32(p[6]) 
    sc_r         = np.float32(p[7])
    sc_g         = np.float32(p[8]) 
    sc_b         = np.float32(p[9])  
    shininess    = int(p[10])
    k_refl       = np.float32(p[11])
    ac           = Color(ac_r, ac_g, ac_b)
    dc           = Color(dc_r, dc_g, dc_b)
    sc           = Color(sc_r, sc_g, sc_b)
    self.materials[material_id] = Material(shininess, ac, dc, sc, k_refl)
