# ------------------------------------------------------------------------------
# Author       : Oystein Gjermundnes
# Organization : Norwegian University of Science and Technology (NTNU)
#                http://www.iet.ntnu.no/en
# Course       : Design of digital systems 1 (DDS1)
# Year         : 2015
# Fork         : git clone git@bitbucket.org:oysteingjermundnes/dds1.git
# Description  : Raytracer model written in python
# License      : This is free and unencumbered software released into the 
#                public domain (UNLICENSE)
# ------------------------------------------------------------------------------

from rayman.vector import *

################################################################################
## Viewing coordinate reference frame class.
#
# This class contains vectors defining the uvn Viewing-Coordinate
# Reference Frame. 
#
# Abbreviations used within this class:
#   - V: View up vector
#   - N: View plane normal vector
#   - n, u, v: Unit vectors representing the uvn viewing-coordinate reference
#     frame
#
# Relevant theory:
#   - [1] pp352-353: The View-Up Vector, The uvn Viewing-Coordinate Reference Frame
#   - [2] pp545-547: Shooting rays
################################################################################
class uvnFrame:
  ##############################################################################
  ## Constructor
  #
  # @param self        Pointer to the uvnFrame.
  # @param cam_pos     vector.Vector representing the camera position.
  # @param cam_target  vector.Vector representing the point the camera is targeting.
  ##############################################################################
  def __init__(self, cam_pos, cam_target):

    ## Set up the uvn Viewing-coordinate Reference Frame ([1] p353)
    ## Viewing plane normal vector
    N     = cam_pos.sub(cam_target)

    ## View-up vector. (Will be automatically made perpendicular to N)               
    V     = Vector(0,1,0)           
                                    

    # V must be perpendicular to N => Dot product must be 0
    # We know N =(x_n, y_n, z_n) and V = (x_v, y_v, z_v),
    # but we need to adjust V so that it is perpendicular to N.
    # We can do that by recomputing x_v as follows:
    # 
    # x_v*x_n + y_v*y_n + z_v*z_n = 0
    # x_v = (-1/x_n)(y_v*y_n + z_v*z_n)

    V_adjusted   = V
    if(V_adjusted.dot(N) != 0.0):
      V_adjusted.x = (-1.0/N.x)*(V.y*N.y + V.z*N.z)

    if(V_adjusted.dot(N) != 0.0):
      print("WARNING: V is not perpendicular to N")
      print("V_adjusted: %s" % V_adjusted.str())
      print("N: %s" % N.str())
      print("V_adjusted*N is: %d" % V_adjusted.dot(N))

    ## View-up vector
    #self.V = V_adjusted
    ## Viewing plane normal vector
    #self.N = N
    ## Normalized normal vector
    self.n = N.normalize()
    ## "Horizontal" unit vector for camera movement
    self.u = V.cross(self.n).normalize()
    ## "Vertical" unit vector for camera movement
    self.v = self.n.cross(self.u)

  def __str__(self):
    return 'u: {}\nv: {}\nn: {}'.format(self.u, self.v, self.n)

