# ------------------------------------------------------------------------------
# Author       : Oystein Gjermundnes
# Organization : Norwegian University of Science and Technology (NTNU)
#                http://www.iet.ntnu.no/en
# Course       : Design of digital systems 1 (DDS1)
# Year         : 2015
# Fork         : git clone git@bitbucket.org:oysteingjermundnes/dds1.git
# Description  : Raytracer model written in python
# License      : This is free and unencumbered software released into the 
#                public domain (UNLICENSE)
# ------------------------------------------------------------------------------

import numpy as np

################################################################################
## Vector class
################################################################################
class Vector:
 
  ##############################################################################
  ## Constructor
  #
  # @param self  Pointer to the Vector.
  # @param x,y,z Coordinates of the Vector.
  ##############################################################################
  def __init__(self, x, y, z):
    ## x-coordinate
    self.x = np.float32(x)
    ## y-coordinate
    self.y = np.float32(y)
    ## z-coordinate
    self.z = np.float32(z)

  ##############################################################################
  ## Normalize
  #
  # Make the vector a vector with length 1.
  #
  # @param self Pointer to the Vector.
  # @return     Vector of length 1.
  ##############################################################################
  def normalize(self):
    return self.mult(1/(self.dot(self)**(0.5))) ## reciprocal sqrt
    #return self.mult(invsqrtapprox_b(self.dot(self)))


  ##############################################################################
  ## Dot product
  #
  # @param self  Pointer to the Vector.
  # @param other Pointer to the other Vector.
  # @return      Dot product between other and self.
  ##############################################################################
  def dot(self, other):
    return self.x*other.x + self.y*other.y + self.z*other.z

  ##############################################################################
  ## Cross product
  #
  # @param self  Pointer to the Vector.
  # @param other Pointer to the other Vector.
  # @return      Cross product between the other and self.
  ##############################################################################
  def cross(self, other):
    return Vector(self.y*other.z - self.z*other.y,
                  self.z*other.x - self.x*other.z,
                  self.x*other.y - self.y*other.x)

  ##############################################################################
  ## Add two vectors.
  #
  # @param self  Pointer to the Vector.
  # @param other Pointer to the other Vector.
  # @return      Sum of other and self.
  ##############################################################################
  def add(self, other):
    return Vector(self.x + other.x,
                  self.y + other.y,
                  self.z + other.z)

  ##############################################################################
  ## Subtract vector
  #
  # @param self  Pointer to the Vector.
  # @param other Pointer to the other Vector.
  # @return      Difference between self and other.
  ##############################################################################
  def sub(self, other):
    return Vector(self.x - other.x,
                  self.y - other.y,
                  self.z - other.z)

  ##############################################################################
  ## Multiply with scalar
  #
  # @param self Pointer to the Vector.
  # @param k    Scalar to multiply every component with.
  # @return     Vector scaled with a factor k.
  ##############################################################################
  def mult(self, k):
    k = np.float32(k)
    return Vector(k*self.x, k*self.y, k*self.z)

  ##############################################################################
  ## Flip the vector in the other direction.
  #
  # @param self Pointer to the Vector.
  # @return     Vector that points in the oposite direction to the input vector.
  ##############################################################################
  def neg(self):
    return Vector(-self.x, -self.y, -self.z)

  ##############################################################################
  ## String representation of the Vector
  #
  # @param self Pointer to the Vector.
  ##############################################################################
  def __str__(self):
    return "Vector(%f, %f, %f)" % (self.x, self.y, self.z)
