module dual_core_array_tb
  ();

  import types_pkg::*;

  function string get_localtime();    
    int fd;
    string localtime;
    void'($dual_core_array("date +%Y-%m-%d_%H.%M.%S > localtime")); // temp file
    fd = $fopen("localtime", "r");
    void'($fscanf(fd,"%s",localtime));
    $fclose(fd);
    void'($dual_core_array("rm localtime")); // delete file
    return localtime;
  endfunction


  // Control
  logic                                reset_n;
  logic                                clk;

  // Inputs
  logic [REQUEST_BITS            -1:0] inp_request;
  logic                                inp_valid;
  logic                                inp_ready;

  // Outputs
  logic [RESULT_BITS             -1:0] out_result;
  logic                                out_valid;
  logic                                out_ready;

  // Object buffer init
  logic                                in_init_load;
  logic [OBJ_BUFFER_ADDR_BITS    -1:0] in_init_addr;
  logic [WORD_BITS               -1:0] in_init_data;

  // Debug
  integer                              active_threads;

  // localparam NUM_REQS = 10_000;
  localparam NUM_REQS = 307_200;
  localparam REQCYCLE = 2;
  localparam NUM_DCORES = 32;


  dual_core_array #(
    .NUM_DCORES(NUM_DCORES)
  ) U0 (
    .*
  );

  // bind ray_core ray_core_property U0 (.*);

  initial begin
    clk = 0;
    reset_n = 0;
    inp_ready = 1;
    #4 reset_n = 1;
  end;

  always
    #1 clk = ~clk;


  //*****************************************************************************
  //
  // Load pregenerated object buffer and init OBUFS with that data
  //
  //*****************************************************************************

  // This does not load the last element! But in this test, it does not matter
  reg [WORD_BITS -1:0] objbuf [0:2**OBJ_BUFFER_ADDR_BITS -1];
  logic [OBJ_BUFFER_ADDR_BITS    -1:0] obj_addr;
  logic                                obj_buf_loaded;
  assign obj_buf_loaded = &obj_addr;
  assign in_init_load = !obj_buf_loaded;
  assign in_init_addr = obj_addr;
  assign in_init_data = objbuf[obj_addr];

  initial begin
    $readmemh("objbuffer.dat", objbuf);
    // $readmemh("objbuffer_lowres.dat", objbuf);
  end

  always_ff @(posedge clk or negedge reset_n) begin
    if (!reset_n) begin
      obj_addr <= 0;
    end
    else if (!obj_buf_loaded) begin
      obj_addr <= obj_addr + 1;
    end
  end


  //*****************************************************************************
  //
  // Save responses from core
  //
  //*****************************************************************************
  int resp_addr = 0;
  reg [RESULT_BITS -1:0] responses [0:NUM_REQS-1];

  always @(posedge clk) begin
    if (inp_ready & out_valid) begin
      responses[resp_addr] <= out_result;
      resp_addr <= resp_addr + 1;
    end
  end


  //*****************************************************************************
  //
  // Load pregenerated requests and send to core
  //
  //*****************************************************************************

  reg [REQUEST_BITS -1:0] requests [0:NUM_REQS-1];

  initial begin
    $readmemb("requests.dat", requests);
    // $readmemb("requests_lowres.dat", requests);
  end

  int cycle_counter;
  always_ff @(posedge clk or negedge reset_n) begin
    if (!reset_n) begin
      cycle_counter <= 0;
    end
    else begin
      cycle_counter <= cycle_counter + 1;
    end
  end

  int req_addr;

  assign inp_request = requests[req_addr];

  assign inp_valid = (req_addr < NUM_REQS) & (req_addr < (cycle_counter/REQCYCLE)) & obj_buf_loaded;

  always_ff @(posedge clk or negedge reset_n) begin
    if (!reset_n) begin
      req_addr <= 0;
    end
    else if (inp_valid & out_ready) begin
      req_addr <= req_addr + 1;
    end
  end


  //*****************************************************************************
  //
  // Stats
  //
  //*****************************************************************************

  integer                              fh;
  integer                              outstanding_requests;
  string                               filename;
  assign outstanding_requests = req_addr - resp_addr - (req_addr == NUM_REQS);
  initial begin
    $sformat(filename, "results/%s.txt", get_localtime()); // + NUM_DCORES
    fh = $fopen(filename);
    // fh = $fopen("active_threads.txt");
    $fwrite(fh, "%d %d\n", NUM_DCORES, REQCYCLE);
    $fmonitor(fh, "%d %d %d", $time, outstanding_requests, active_threads);
  end

  //*****************************************************************************
  //
  // Save results and stop simulation when all requests has given a response
  //
  //*****************************************************************************

  always_comb begin
    if (resp_addr == NUM_REQS) begin
      $writememb("responses_calc.dat", responses);
      $fwrite(fh, "%d %d %d", $time, outstanding_requests+1, active_threads);
      $fclose(fh);
      $stop;
    end
  end


endmodule